import helpers
import numpy as np
import pytest

import toughio


def test_flow():
    def write_read(x):
        writer_kws = {"file_format": "toughreact-flow"}
        reader_kws = {"file_format": "toughreact-flow"}

        return helpers.write_read(
            "flow.inp",
            x,
            toughio.write_input,
            toughio.read_input,
            writer_kws,
            reader_kws,
        )

    parameters_ref = {
        "rocks": {
            helpers.random_string(5): {
                "tortuosity": -np.random.rand(),
                "porosity_crit": np.random.rand(),
                "tortuosity_exponent": np.random.rand(),
            },
            helpers.random_string(5): {
                "relative_permeability": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(3),
                },
                "capillarity": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(4),
                },
            },
            helpers.random_string(5): {
                "react_tp": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(3),
                },
            },
            helpers.random_string(5): {
                "react_hcplaw": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(4),
                },
            },
            helpers.random_string(5): {
                "react_tp": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(3),
                },
                "react_hcplaw": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(4),
                },
            },
            helpers.random_string(5): {
                "relative_permeability": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(3),
                },
                "react_tp": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(3),
                },
            },
            helpers.random_string(5): {
                "capillarity": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(4),
                },
                "react_hcplaw": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(4),
                },
            },
            helpers.random_string(5): {
                "relative_permeability": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(3),
                },
                "capillarity": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(4),
                },
                "react_tp": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(3),
                },
                "react_hcplaw": {
                    "id": np.random.randint(10),
                    "parameters": np.random.rand(4),
                },
            },
        },
        "react": {
            "options": {k + 1: v for k, v in enumerate(np.random.randint(10, size=25))},
            "output": {
                "format": np.random.randint(10),
                "shape": np.random.randint(100, size=3),
            },
            "poiseuille": {
                "start": np.random.rand(2),
                "end": np.random.rand(2),
                "aperture": np.random.rand(),
            },
        },
        "options": {
            "react_wdata": [
                helpers.random_string(5) for _ in range(np.random.randint(10) + 1)
            ],
        },
        "initial_conditions": {
            helpers.random_string(5): {
                "values": np.random.rand(4),
                "permeability": np.random.rand(3),
            },
        },
        "generators": [
            {
                "label": helpers.random_string(5),
                "type": helpers.random_string(4),
                "rates": np.random.rand(),
                "specific_enthalpy": np.random.rand(),
            },
            {
                "label": helpers.random_string(5),
                "type": helpers.random_string(4),
                "times": np.random.rand(10),
                "rates": np.random.rand(10),
                "specific_enthalpy": np.random.rand(10),
            },
            {
                "label": helpers.random_string(5),
                "type": helpers.random_string(4),
                "rates": np.random.rand(),
                "specific_enthalpy": np.random.rand(),
                "conductivity_times": np.random.rand(5),
                "conductivity_factors": np.random.rand(5),
            },
            {
                "label": helpers.random_string(5),
                "type": helpers.random_string(4),
                "times": np.random.rand(10),
                "rates": np.random.rand(10),
                "specific_enthalpy": np.random.rand(10),
                "conductivity_times": np.random.rand(9),
                "conductivity_factors": np.random.rand(9),
            },
        ],
    }
    parameters = write_read(parameters_ref)

    assert helpers.allclose(parameters_ref, parameters, atol=1.0e-4)


@pytest.mark.parametrize(
    "verbose, mopr_10, mopr_11",
    [
        (True, 0, 0),
        (True, 2, 0),
        (True, 0, 1),
        (True, 0, 2),
        (False, 0, 0),
        (False, 2, 0),
        (False, 0, 1),
        (False, 0, 2),
    ],
)
def test_solute(verbose, mopr_10, mopr_11):
    def write_read(x):
        writer_kws = {
            "file_format": "toughreact-solute",
            "verbose": verbose,
            "mopr_10": mopr_10,
            "mopr_11": mopr_11,
        }
        reader_kws = {
            "file_format": "toughreact-solute",
            "mopr_11": mopr_11,
        }

        return helpers.write_read(
            "solute.inp",
            x,
            toughio.write_input,
            toughio.read_input,
            writer_kws,
            reader_kws,
        )

    parameters_ref = {
        "title": helpers.random_string(80),
        "options": {
            "sl_min": np.random.rand(),
            "rcour": np.random.rand(),
            "ionic_strength_max": np.random.rand(),
            "mineral_gas_factor": np.random.rand(),
            "w_time": np.random.rand(),
            "w_upstream": np.random.rand(),
            "aqueous_diffusion_coefficient": np.random.rand(),
            "molecular_diffusion_coefficient": np.random.rand(),
            "n_iteration_tr": np.random.randint(100),
            "eps_tr": np.random.rand(),
            "n_iteration_ch": np.random.randint(100),
            "eps_ch": np.random.rand(),
            "eps_mb": np.random.rand(),
            "eps_dc": np.random.rand(),
            "eps_dr": np.random.rand(),
            "n_cycle_print": np.random.randint(100),
        },
        "flags": {
            "iteration_scheme": np.random.randint(10),
            "reactive_surface_area": np.random.randint(10),
            "solver": np.random.randint(10),
            "n_subiteration": np.random.randint(10),
            "gas_transport": np.random.randint(10),
            "verbosity": np.random.randint(10),
            "feedback": np.random.randint(10),
            "coupling": np.random.randint(10),
            "aqueous_concentration_unit": np.random.randint(10),
            "mineral_unit": np.random.randint(10),
            "gas_concentration_unit": np.random.randint(10),
        },
        "files": {
            "thermodynamic_input": helpers.random_string(20),
            "iteration_output": helpers.random_string(20),
            "plot_output": helpers.random_string(20),
            "solid_output": helpers.random_string(20),
            "gas_output": helpers.random_string(20),
            "time_output": helpers.random_string(20),
        },
        "output": {
            "elements": [
                helpers.random_string(5) for _ in range(np.random.randint(10) + 1)
            ],
            "components": [
                helpers.random_string(20) for _ in range(np.random.randint(10) + 1)
            ],
            "minerals": np.random.randint(10, size=np.random.randint(10) + 1),
            "aqueous_species": [
                helpers.random_string(20) for _ in range(np.random.randint(10) + 1)
            ],
            "surface_complexes": np.random.randint(10, size=np.random.randint(10) + 1),
            "exchange_species": [
                helpers.random_string(20) for _ in range(np.random.randint(10) + 1)
            ],
        },
        "default": {
            "initial_water": np.random.randint(10),
            "injection_water": np.random.randint(10),
            "mineral": np.random.randint(10),
            "initial_gas": np.random.randint(10),
            "adsorption": np.random.randint(10),
            "cation_exchange": np.random.randint(10),
            "permeability_porosity": np.random.randint(10),
            "linear_kd": np.random.randint(10),
            "injection_gas": np.random.randint(10),
        },
        "zones": {
            helpers.random_string(5): {
                "nseq": np.random.randint(10),
                "nadd": np.random.randint(10),
                "initial_water": np.random.randint(10),
                "injection_water": np.random.randint(10),
                "mineral": np.random.randint(10),
                "initial_gas": np.random.randint(10),
                "adsorption": np.random.randint(10),
                "cation_exchange": np.random.randint(10),
                "permeability_porosity": np.random.randint(10),
                "linear_kd": np.random.randint(10),
                "injection_gas": np.random.randint(10),
            }
            for _ in range(np.random.randint(10) + 1)
        },
        "end_comments": [
            helpers.random_string(80) for _ in range(np.random.randint(5) + 2)
        ],
    }

    if mopr_10 == 2:
        parameters_ref["options"].update(
            {
                "n_iteration_1": np.random.randint(100),
                "n_iteration_2": np.random.randint(100),
                "n_iteration_3": np.random.randint(100),
                "n_iteration_4": np.random.randint(100),
                "t_increase_factor_1": np.random.rand(),
                "t_increase_factor_2": np.random.rand(),
                "t_increase_factor_3": np.random.rand(),
                "t_reduce_factor_1": np.random.rand(),
                "t_reduce_factor_2": np.random.rand(),
                "t_reduce_factor_3": np.random.rand(),
            }
        )

    if mopr_11 == 1:
        parameters_ref["default"]["sedimentation_velocity"] = np.random.rand()
        for zone in parameters_ref["zones"].values():
            zone["sedimentation_velocity"] = np.random.rand()

    elif mopr_11 == 2:
        parameters_ref["default"]["element"] = np.random.randint(10)
        for zone in parameters_ref["zones"].values():
            zone["element"] = np.random.randint(10)

    parameters = write_read(parameters_ref)

    assert helpers.allclose(parameters_ref, parameters, atol=1.0e-3)


@pytest.mark.parametrize("verbose", [True, False])
def test_chemical(verbose):
    def write_read(x):
        writer_kws = {
            "verbose": verbose,
            "file_format": "toughreact-chemical",
        }
        reader_kws = {
            "file_format": "toughreact-chemical",
        }

        return helpers.write_read(
            "chemical.inp",
            x,
            toughio.write_input,
            toughio.read_input,
            writer_kws,
            reader_kws,
        )

    # Number of surface primary species
    nsurfs = np.random.randint(5) + 1

    parameters_ref = {
        "title": helpers.random_string(80),
        "primary_species": [
            {
                "name": helpers.random_string(20),
                "transport": np.random.randint(2),
            }
            for _ in range(np.random.randint(5) + 1)
        ],
        "aqueous_kinetics": [
            {
                "species": [
                    {
                        "name": helpers.random_string(20),
                        "stoichiometric_coeff": np.random.rand(),
                    }
                    for _ in range(np.random.randint(5) + 1)
                ],
                "reaction_affinity": {
                    "id": np.random.randint(5),
                    "cf": np.random.rand(),
                    "logK": np.random.rand(),
                },
                "id": np.random.randint(4),
                "n_mechanism": np.random.randint(10),
                "product": [
                    {
                        "specie": helpers.random_string(20),
                        "flag": np.random.randint(3),
                        "power": np.random.rand(),
                    }
                    for _ in range(np.random.randint(5) + 1)
                ],
                "monod": [
                    {
                        "specie": helpers.random_string(20),
                        "flag": np.random.randint(3),
                        "half_saturation": np.random.rand(),
                    }
                    for _ in range(np.random.randint(5) + 1)
                ],
                "inhibition": [
                    {
                        "specie": helpers.random_string(20),
                        "flag": np.random.randint(3),
                        "constant": np.random.rand(),
                    }
                    for _ in range(np.random.randint(5) + 1)
                ],
            }
            for _ in range(2)
        ],
        "aqueous_species": [
            helpers.random_string(20) for _ in range(np.random.randint(5) + 1)
        ],
        "minerals": [
            {
                "name": helpers.random_string(20),
                "type": 0,
                "kinetic_constraint": 0,
                "solid_solution": np.random.randint(3),
                "precipitation_dry": np.random.randint(3),
                "gap": np.random.rand(),
                "temp1": np.random.rand(),
                "temp2": np.random.rand(),
            },
            {
                "name": helpers.random_string(20),
                "type": 1,
                "kinetic_constraint": 1,
                "solid_solution": np.random.randint(3),
                "precipitation_dry": np.random.randint(3),
                "dissolution": {
                    "k25": np.random.rand(),
                    "rate_ph_dependence": 1,
                    "eta": np.random.rand(),
                    "theta": np.random.rand(),
                    "activation_energy": np.random.rand(),
                    "a": np.random.rand(),
                    "b": np.random.rand(),
                    "c": np.random.rand(),
                    "ph1": np.random.rand(),
                    "slope1": np.random.rand(),
                    "ph2": np.random.rand(),
                    "slope2": np.random.rand(),
                },
            },
            {
                "name": helpers.random_string(20),
                "type": 1,
                "kinetic_constraint": 2,
                "solid_solution": np.random.randint(3),
                "precipitation_dry": np.random.randint(3),
                "precipitation": {
                    "k25": np.random.rand(),
                    "rate_ph_dependence": 2,
                    "eta": np.random.rand(),
                    "theta": np.random.rand(),
                    "activation_energy": np.random.rand(),
                    "a": np.random.rand(),
                    "b": np.random.rand(),
                    "c": np.random.rand(),
                    "volume_fraction_ini": np.random.rand(),
                    "id": np.random.randint(2),
                    "extra_mechanisms": [
                        {
                            "ki": np.random.rand(),
                            "activation_energy": np.random.rand(),
                            "species": [
                                {
                                    "name": helpers.random_string(20),
                                    "power": np.random.rand(),
                                }
                                for _ in range(np.random.randint(5) + 1)
                            ],
                        }
                        for _ in range(np.random.randint(5) + 1)
                    ],
                },
                "gap": np.random.rand(),
                "temp1": np.random.rand(),
                "temp2": np.random.rand(),
            },
            {
                "name": helpers.random_string(20),
                "type": 1,
                "kinetic_constraint": 3,
                "solid_solution": np.random.randint(3),
                "precipitation_dry": np.random.randint(3),
                "dissolution": {
                    "k25": np.random.rand(),
                    "rate_ph_dependence": 0,
                    "eta": np.random.rand(),
                    "theta": np.random.rand(),
                    "activation_energy": np.random.rand(),
                    "a": np.random.rand(),
                    "b": np.random.rand(),
                    "c": np.random.rand(),
                },
                "precipitation": {
                    "k25": np.random.rand(),
                    "rate_ph_dependence": 0,
                    "eta": np.random.rand(),
                    "theta": np.random.rand(),
                    "activation_energy": np.random.rand(),
                    "a": np.random.rand(),
                    "b": np.random.rand(),
                    "c": np.random.rand(),
                    "volume_fraction_ini": np.random.rand(),
                    "id": np.random.randint(2),
                },
                "gap": np.random.rand(),
                "temp1": np.random.rand(),
                "temp2": np.random.rand(),
            },
        ],
        "gaseous_species": [
            {
                "name": helpers.random_string(20),
                "fugacity": np.random.randint(2),
            }
            for _ in range(np.random.randint(5) + 1)
        ],
        "surface_complexes": [
            helpers.random_string(20) for _ in range(np.random.randint(5) + 1)
        ],
        "kd_decay": [
            {
                "name": helpers.random_string(20),
                "decay_constant": np.random.rand(),
                "a": np.random.rand(),
                "b": np.random.rand(),
            }
            for _ in range(np.random.randint(5) + 1)
        ],
        "exchanged_species": [
            {
                "name": helpers.random_string(20),
                "reference": bool(np.random.randint(2)),
                "type": np.random.randint(4),
                "site_coeffs": np.random.rand(3),
            }
            for _ in range(np.random.randint(5) + 1)
        ],
        "exchange_sites_id": np.random.randint(3),
        "zones": {
            "minerals": [
                {
                    "species": [
                        {
                            "name": helpers.random_string(20),
                            "volume_fraction_ini": np.random.rand(),
                            "flag": 0,
                        },
                        {
                            "name": helpers.random_string(20),
                            "volume_fraction_ini": np.random.rand(),
                            "flag": 1,
                            "radius": np.random.rand(),
                            "area_ini": np.random.rand(),
                            "area_unit": np.random.randint(5),
                        },
                    ],
                },
                {
                    "rock": helpers.random_string(5),
                    "species": [
                        {
                            "name": helpers.random_string(20),
                            "volume_fraction_ini": np.random.rand(),
                            "flag": 0,
                        }
                        for _ in range(np.random.randint(5) + 1)
                    ],
                },
            ],
            "permeability_porosity": [
                {
                    "id": np.random.randint(7),
                    "a": np.random.rand(),
                    "b": np.random.rand(),
                }
                for _ in range(np.random.randint(5) + 1)
            ],
            "adsorption": [
                {
                    "flag": np.random.randint(2),
                    "species": [
                        {
                            "name": helpers.random_string(20),
                            "area_unit": np.random.randint(3),
                            "area": np.random.rand(),
                        }
                        for _ in range(nsurfs)
                    ],
                }
                for _ in range(np.random.randint(5) + 1)
            ],
            "linear_kd": [
                [
                    {
                        "name": helpers.random_string(20),
                        "solid_density": np.random.rand(),
                        "value": np.random.rand(),
                    }
                    for _ in range(np.random.randint(5) + 1)
                ]
                for _ in range(np.random.randint(5) + 1)
            ],
            "cation_exchange": [
                np.random.rand(5) for _ in range(np.random.randint(5) + 1)
            ],
        },
        "end_comments": [
            helpers.random_string(80) for _ in range(np.random.randint(5) + 2)
        ],
    }

    # Primary species
    for _ in range(nsurfs):
        tmp = {
            "name": helpers.random_string(20),
            "transport": 2,
            "mineral_name": helpers.random_string(20),
            "sorption_density": np.random.rand(),
            "adsorption_id": np.random.randint(5),
        }
        if tmp["adsorption_id"] == 1:
            tmp["capacitance"] = np.random.rand()

        parameters_ref["primary_species"].append(tmp)

    # Aqueous kinetics
    parameters_ref["aqueous_kinetics"][0]["rate"] = np.random.rand()
    parameters_ref["aqueous_kinetics"][1]["rate"] = {
        "k25": np.random.rand(),
        "Ea": np.random.rand(),
    }

    # Initial and injection water zones
    for key in ["initial_waters", "injection_waters"]:
        parameters_ref["zones"][key] = [
            {
                "temperature": np.random.rand(),
                "pressure": np.random.rand(),
                "species": [
                    {
                        "name": helpers.random_string(20),
                        "flag": np.random.randint(5),
                        "guess": np.random.rand(),
                        "ctot": np.random.rand(),
                        "log_fugacity": np.random.rand(),
                        "nameq": helpers.random_string(10),
                    }
                    for _ in range(np.random.randint(5) + 1)
                ],
            },
            {
                "temperature": np.random.rand(),
                "pressure": np.random.rand(),
                "rock": helpers.random_string(5),
                "species": [
                    {
                        "name": helpers.random_string(20),
                        "flag": np.random.randint(5),
                        "guess": np.random.rand(),
                        "ctot": np.random.rand(),
                        "log_fugacity": np.random.rand(),
                        "nameq": helpers.random_string(10),
                    }
                    for _ in range(np.random.randint(5) + 1)
                ],
            },
        ]

    # Initial and injection gas zones
    for key in ["initial_gases", "injection_gases"]:
        pp = "partial_pressure" if key == "initial_gases" else "mole_fraction"
        parameters_ref["zones"][key] = [
            [
                {
                    "name": helpers.random_string(20),
                    pp: np.random.rand(),
                }
                for _ in range(np.random.randint(5) + 1)
            ]
            for _ in range(np.random.randint(5) + 1)
        ]

    parameters = write_read(parameters_ref)

    assert helpers.allclose(parameters_ref, parameters, atol=1.0e-3)



import numpy as np

from ...._common import open_file

__all__ = [
    "read",
]


def read(filename):
    """
    Read CSV table file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.

    Returns
    -------
    dict
        Table data.

    Note
    ----
    Also supports iTOUGH2 .foft files.

    """
    with open_file(filename, "r") as f:
        # Read headers
        line = f.readline().strip()
        sep = "," if "," in line else None

        headers = []
        while True:
            hdr = [x.replace('"', "").strip() for x in line.split(sep)]

            try:
                _ = float(hdr[0])
                break

            except ValueError:
                headers.append(hdr)
                line = f.readline().strip()

        headers = [" ".join(hdr) for hdr in zip(*headers)]

        # Read data
        data = []
        while line:
            data.append([x.replace('"', "").strip() for x in line.split(sep)])
            line = f.readline().strip()

        out = {}
        for header, X in zip(headers, np.transpose(data)):
            # Remove extra whitespaces
            header = " ".join(header.split())

            if header in {"KCYC", "ROCK"}:
                out[header] = np.array([int(x) for x in X])

            elif header.startswith("ELEM"):
                label_length = max(len(x) for x in X)
                fmt = f"{{:>{label_length}}}"
                out[header] = np.array([fmt.format(x) for x in X])

            else:
                out[header] = np.array([float(x) for x in X])

        return out



from ..._common import filetype_from_filename, register_format

__all__ = [
    "register",
    "read",
]


_extension_to_filetype = {}
_reader_map = {}
_writer_map = {}


def register(file_format, extensions, reader, writer=None):
    """
    Register a new table format.

    Parameters
    ----------
    file_format : str
        File format to register.
    extensions : array_like
        List of extensions to associate to the new format.
    reader : callable
        Read function.
    writer : callable or None, optional, default None
        Write function.

    """
    register_format(
        fmt=file_format,
        ext_to_fmt=_extension_to_filetype,
        reader_map=_reader_map,
        writer_map=_writer_map,
        extensions=extensions,
        reader=reader,
        writer=writer,
    )


def read(filename, file_format=None, **kwargs):
    """
    Read table file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.
    file_format : str ('column', 'csv', 'tecplot') or None, optional, default None
        Input file format.

    Returns
    -------
    dict
        Table data.

    """
    fmt = (
        file_format
        if file_format
        else filetype_from_filename(filename, _extension_to_filetype, "csv")
    )

    return _reader_map[fmt](filename, **kwargs)



from functools import partial

import numpy as np

from ...._common import open_file
from ..._common import read_record
from .._common import to_output

__all__ = [
    "read",
]


def read(filename, file_type, labels_order=None):
    """
    Read standard TOUGH output file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.
    file_type : str
        Input file type.
    labels_order : list of array_like
        List of labels. If None, output will be assumed ordered.

    Returns
    -------
    namedtuple or list of namedtuple
        namedtuple (type, format, time, labels, data) or list of namedtuple for each time step.

    """
    with open_file(filename, "r") as f:
        headers, times, variables = _read_table(f, file_type)

        # Postprocess labels
        labels = [v[0].lstrip() for v in variables[0]]

        if file_type == "element":
            label_length = max(len(label) for label in labels)
            fmt = f"{{:>{label_length}}}"
            labels = [fmt.format(label) for label in labels]

        else:
            # Find end of first element processing backward
            labels = [label[::-1] for label in labels]
            tmp = max(labels, key=len)

            while True:
                iend = tmp.index(" ")

                if iend < 2:
                    tmp = f"{tmp[:iend]}0{tmp[iend + 1:]}"

                else:
                    break

            iend += len(tmp[iend:]) - len(tmp[iend:].lstrip())

            # Split connection names
            labels1 = [label[iend:].rstrip() for label in labels]
            labels2 = [label[:iend].rstrip() for label in labels]

            # Find label length
            len1 = max(len(label) for label in labels1)
            len2 = max(len(label) for label in labels2)
            label_length = max(len1, len2)

            # Correct element names given label length
            fmt = f"{{:<{label_length}}}"
            labels1 = [fmt.format(label) for label in labels1]
            labels2 = [fmt.format(label) for label in labels2]
            labels = [[l1[::-1], l2[::-1]] for l1, l2 in zip(labels1, labels2)]

        ilab = 1 if file_type == "element" else 2
        headers = headers[ilab + 1 :]
        labels = [labels.copy() for _ in variables]
        variables = np.array([[v[2:] for v in variable] for variable in variables])

    return to_output(file_type, labels_order, headers, times, labels, variables)


def _read_table(f, file_type):
    """Read data table for current time step."""
    labels_key = "ELEM." if file_type == "element" else "ELEM1"

    first = True
    times, variables = [], []
    for line in f:
        line = line.strip()

        # Look for "TOTAL TIME"
        if line.startswith("TOTAL TIME"):
            # Read time step in following line
            line = next(f).strip()
            times.append(float(line.split()[0]))
            variables.append([])

            # Look for "ELEM." or "ELEM1"
            while True:
                try:
                    line = next(f).strip()
                    if line.startswith(labels_key):
                        break

                except StopIteration:
                    raise ValueError(f"No data related to {file_type}s found.")

            # Read headers
            headers = line.split()

            # Read units
            line = next(f)
            nwsp = line.index(
                line.strip()[0]
            )  # Index of first non whitespace character

            # Look for next non-empty line
            while True:
                line = next(f)
                if line.strip():
                    break

            # Loop until end of output block
            while True:
                if line[:nwsp].strip() and not line.strip().startswith("ELEM"):
                    if first:
                        # Find first floating point
                        x = line.split()[-headers[::-1].index("INDEX")]

                        # Find end of label(s)
                        tmp = line[: line.index(x)].rstrip()
                        tmp = tmp[::-1]
                        tmp = tmp[: line.index(" ") : -1].rstrip()
                        iend = len(tmp)

                    tmp = [line[:iend]]
                    line = line[iend:]

                    if first:
                        try:
                            # Set line parser and try parsing first line
                            reader = lambda line: [to_float(x) for x in line.split()]
                            _ = reader(line)

                        except ValueError:
                            # Determine number of characters for index
                            idx = line.replace("-", " ").split()[0]
                            nidx = line.index(idx) + len(idx)
                            ifmt = f"{nidx}s"

                            # Determine number of characters between two Es
                            i1 = line.find("E")
                            i2 = line.find("E", i1 + 1)

                            # Initialize data format
                            fmt = [ifmt]
                            if i2 >= 0:
                                di = i2 - i1
                                dfmt = f"{di}.{di - 7}e"
                                fmt += 20 * [dfmt]  # Read 20 data columns at most

                            else:
                                fmt += ["12.5e"]

                            fmt = ",".join(fmt)

                            # Set line parser
                            reader = partial(read_record, fmt=fmt)

                        finally:
                            first = False

                    tmp += reader(line)
                    variables[-1].append([x for x in tmp if x is not None])

                line = next(f)
                if line[1:].startswith("@@@@@"):
                    break

    return headers, times, variables


def to_float(x):
    """Return np.nan if x cannot be converted."""
    from ..._common import to_float as _to_float

    try:
        return _to_float(x)

    except ValueError:
        return np.nan



import numpy as np


def read_record(data, fmt):
    """Parse string to data given format."""
    token_to_type = {
        "s": str,
        "S": str,
        "d": int,
        "f": to_float,
        "e": to_float,
    }

    i = 0
    out = []
    for token in fmt.split(","):
        n = int(token[:-1].split(".")[0])
        tmp = data[i : i + n]
        tmp = tmp if token[-1] == "S" else tmp.strip()
        out.append(token_to_type[token[-1]](tmp) if tmp else None)
        i += n

    return out


def write_record(data, fmt, space_between_values=False, multi=False):
    """Return a list of record strings given format."""
    if not multi:
        data = [to_str(d, f, space_between_values) for d, f in zip(data, fmt)]
        out = [f"{''.join(data):80}\n"]

    else:
        n = len(data)
        ncol = len(fmt)
        data = [
            data[ncol * i : min(ncol * i + ncol, n)]
            for i in range(int(np.ceil(n / ncol)))
        ]

        out = []
        for d in data:
            d = [to_str(dd, f, space_between_values) for dd, f in zip(d, fmt)]
            out += [f"{''.join(d):80}\n"]

    return out


def to_float(s):
    """Convert variable string to float."""
    try:
        return float(s.replace("d", "e"))

    except ValueError:
        # It's probably something like "0.0001-001"
        significand, exponent = s[:-4], s[-4:]

        return float(f"{significand}e{exponent}")


def to_str(x, fmt, space_between_values=False):
    """Convert variable to string."""
    x = "" if x is None else x

    if not isinstance(x, str):
        if "f" in fmt:
            tmp = str(float(x))

            n = int(fmt[3:].split("f")[0])
            fmt = f"{{:>{n}}}"

            if space_between_values:
                n -= 1

            if len(tmp) > n or "e" in tmp:
                tmp = (
                    tmp[:n]
                    if 1.0 <= abs(x) < 10.0
                    else format(scientific_notation(x, n))
                )

            return fmt.format(tmp)

        else:
            return fmt.format(x)

    else:
        return fmt.replace("g", "").replace("f", "").format(x)


def scientific_notation(x, n):
    """
    Scientific notation with fixed number of characters.

    Note
    ----
    This function maximizes accuracy given a fixed number of characters.

    """
    tmp = np.format_float_scientific(
        x,
        unique=True,
        trim="0",
        exp_digits=0,
        sign=False,
    )
    tmp = tmp.replace("+", "")

    if len(tmp) > n:
        significand, exponent = tmp.split("e")
        significand = significand[: n - len(tmp)]

        return f"{significand}e{exponent}"

    else:
        return tmp



import os

import helpers
import numpy as np
import pytest

import toughio

write_read = lambda output, writer_kws, reader_kws: helpers.write_read(
    "output",
    output,
    toughio.write_output,
    toughio.read_output,
    writer_kws=writer_kws,
    reader_kws=reader_kws,
)


@pytest.mark.parametrize(
    "filename, filename_ref, file_format",
    [
        ("OUTPUT_ELEME.csv", "SAVE.out", "csv"),
        ("OUTPUT_ELEME.tec", "SAVE.out", "tecplot"),
        ("OUTPUT_ELEME_PETRASIM.csv", "SAVE.out", "petrasim"),
        ("OUTPUT.out", "SAVE.out", "tough"),
        ("OUTPUT_6.out", "SAVE_6.out", "tough"),
    ],
)
def test_output_eleme(filename, filename_ref, file_format):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", filename)
    outputs = toughio.read_output(filename)

    filename = os.path.join(this_dir, "support_files", "outputs", filename_ref)
    save = toughio.read_output(filename)

    assert len(outputs) == 5

    times_ref = [
        0.2592000e08,
        0.3155800e08,
        0.1577900e09,
        0.3155800e09,
        0.7889400e09,
    ]
    keys_ref = ["POR", "PRES", "SAT_G", "TEMP", "X", "Y", "Z"]
    for output, time_ref in zip(outputs, times_ref):
        assert time_ref == output.time
        assert (
            save.labels.tolist() == output.labels.tolist()
            if file_format in {"csv", "petrasim", "tough"}
            else len(output.labels) == 0
        )
        if file_format != "tough":
            assert keys_ref == sorted(list(output.data))

    assert helpers.allclose(save.data["X1"], outputs[-1].data["PRES"])
    assert helpers.allclose(save.data["X2"], outputs[-1].data["TEMP"], atol=0.1)


@pytest.mark.parametrize(
    "filename",
    ["OUTPUT_CONNE.csv", "OUTPUT.out", "OUTPUT_6.out"],
)
def test_output_conne(filename):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", filename)
    outputs = toughio.read_output(filename, connection=True)

    times_ref = [
        0.2592000e08,
        0.3155800e08,
        0.1577900e09,
        0.3155800e09,
        0.7889400e09,
    ]
    data_ref = [
        52542.0,
        52475.0,
        51146.0,
        49600.0,
        45623.0,
    ]
    for output, time_ref, data in zip(outputs, times_ref, data_ref):
        assert time_ref == output.time
        assert (
            len(set("".join(labels) for labels in output.labels))
            == output.data["HEAT"].size
        )
        assert helpers.allclose(data, np.abs(output.data["HEAT"]).mean(), atol=1.0)


@pytest.mark.parametrize(
    "output_ref, file_format",
    [
        (helpers.output_eleme, "csv"),
        (helpers.output_eleme[0], "csv"),
        (helpers.output_eleme, "petrasim"),
        (helpers.output_eleme[0], "petrasim"),
        (helpers.output_eleme, "tecplot"),
        (helpers.output_eleme[0], "tecplot"),
        (helpers.output_conne, "csv"),
        (helpers.output_conne[0], "csv"),
    ],
)
def test_output(output_ref, file_format):
    output = write_read(
        output=output_ref,
        writer_kws={"file_format": file_format},
        reader_kws={},
    )

    output_ref = output_ref if isinstance(output_ref, list) else [output_ref]
    for out_ref, out in zip(output_ref, output):
        # Careful here, tecplot format has no label
        helpers.allclose(out, out_ref)


def test_save():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "outputs", "SAVE.out")
    save = toughio.read_output(filename)

    x_ref = [6.35804123e05, 1.42894499e02, 9.91868799e-01]
    assert helpers.allclose(
        x_ref, np.mean([save.data["X1"], save.data["X2"], save.data["X3"]], axis=1)
    )

    assert helpers.allclose(0.01, save.data["porosity"].mean())

    assert "userx" not in save.data



from ...._common import open_file, prune_values
from ...._exceptions import ReadError
from ...._helpers import FileIterator
from ..._common import read_record, to_float
from .._common import read_end_comments

__all__ = [
    "read",
]


def read(filename, mopr_11=0):
    """
    Read TOUGHREACT solute input file.

    Parameters
    ----------
    filename : str
        Input file name.
    mopr_11 : int, optional, default 0
        MOPR(11) value in file 'flow.inp'.

    """
    with open_file(filename, "r") as f:
        out = read_buffer(f, mopr_11)

    return out


def read_buffer(f, mopr_11=0):
    """Read TOUGHREACT solute input file."""
    parameters = {}
    fiter = FileIterator(f)

    try:
        parameters.update(_read_title(fiter))
        parameters.update(_read_options(fiter))
        parameters.update(_read_filenames(fiter))
        parameters["options"].update(_read_weights_coefficients(fiter))
        parameters["options"].update(_read_convergence(fiter))

        output, numbers = _read_output(fiter)
        parameters["options"].update(output["options"])
        parameters["flags"].update(output["flags"])
        parameters["output"] = {}

        elements = _read_elements(fiter, numbers["NWNOD"])
        if elements:
            parameters["output"]["elements"] = elements

        components = _read_indices_names(fiter, numbers["NWCOM"])
        if components:
            parameters["output"]["components"] = components

        minerals = _read_indices_names(fiter, numbers["NWMIN"])
        if minerals:
            parameters["output"]["minerals"] = minerals

        aqueous_species = _read_indices_names(fiter, numbers["NWAQ"])
        if aqueous_species:
            parameters["output"]["aqueous_species"] = aqueous_species

        surface_complexes = _read_indices_names(fiter, numbers["NWADS"])
        if surface_complexes:
            parameters["output"]["surface_complexes"] = surface_complexes

        exchange_species = _read_indices_names(fiter, numbers["NWEXC"])
        if exchange_species:
            parameters["output"]["exchange_species"] = exchange_species

        parameters.update(_read_default(fiter, mopr_11))
        parameters.update(_read_zones(fiter, mopr_11))

        convergence_bounds = _read_convergence_bounds(fiter)
        if convergence_bounds:
            parameters["options"].update(convergence_bounds)

        # Look for file ending
        while fiter.line.strip() != "end":
            try:
                _ = fiter.next()

            except StopIteration:
                break

        # End comments
        end_comments = read_end_comments(fiter)
        if end_comments:
            parameters["end_comments"] = end_comments

    except:
        raise ReadError(f"failed to parse line {fiter.count}.")

    return parameters


def _read_title(f):
    """Read title."""
    line = f.next(skip_empty=True, comments="#").strip()

    return {"title": line}


def _read_options(f):
    """Read options."""
    # Record 2
    line = f.next(skip_empty=True, comments="#").strip()
    data = [int(x) for x in line.split()]
    if len(data) < 8:
        raise ReadError()

    options = {
        "flags": {
            "iteration_scheme": data[0],
            "reactive_surface_area": data[1],
            "solver": data[2],
            "n_subiteration": data[3],
            "gas_transport": data[4],
            "verbosity": data[5],
            "feedback": data[6],
            "coupling": data[7],
            # ITDS_REACT is not used
        }
    }

    # Record 3
    line = f.next(skip_empty=True, comments="#").strip()
    data = [to_float(x) for x in line.split()]
    if len(data) < 4:
        raise ReadError()

    options["options"] = {
        "sl_min": data[0],
        "rcour": data[1],
        "ionic_strength_max": data[2],
        "mineral_gas_factor": data[3],
    }

    return options


def _read_filenames(f):
    """Read file names."""
    filenames = {"files": {}}

    line = f.next(skip_empty=True, comments="#")
    filenames["files"]["thermodynamic_input"] = line[:20].strip()

    for keys in ["iteration", "plot", "solid", "gas", "time"]:
        line = f.next()
        filenames["files"]["{}_output".format(keys)] = line[:20].strip()

    return filenames


def _read_weights_coefficients(f):
    """Read weights and diffusion coefficients."""
    line = f.next(skip_empty=True, comments="#").strip()
    data = [to_float(x) for x in line.split()]
    if len(data) < 4:
        raise ReadError()

    options = {
        "w_time": data[0],
        "w_upstream": data[1],
        "aqueous_diffusion_coefficient": data[2],
        "molecular_diffusion_coefficient": data[3],
    }

    return options


def _read_convergence(f):
    """Read convergence criterion."""
    line = f.next(skip_empty=True, comments="#").strip()
    data = line.split()
    if len(data) < 7:
        raise ReadError()

    options = {
        "n_iteration_tr": int(data[0]),
        "eps_tr": to_float(data[1]),
        "n_iteration_ch": int(data[2]),
        "eps_ch": to_float(data[3]),
        "eps_mb": to_float(data[4]),
        "eps_dc": to_float(data[5]),
        "eps_dr": to_float(data[6]),
    }

    return options


def _read_output(f):
    """Read output control variables."""
    output = {"options": {}, "flags": {}}

    line = f.next(skip_empty=True, comments="#").strip()
    data = [int(x) for x in line.split()]
    if len(data) < 10:
        raise ReadError()

    output = {
        "options": {"n_cycle_print": data[0]},
        "flags": {
            "aqueous_concentration_unit": data[7],
            "mineral_unit": data[8],
            "gas_concentration_unit": data[9],
        },
    }

    numbers = {
        "NWNOD": data[1],
        "NWCOM": data[2],
        "NWMIN": data[3],
        "NWAQ": data[4],
        "NWADS": data[5],
        "NWEXC": data[6],
    }

    return output, numbers


def _read_elements(f, n):
    """Read list of grid blocks."""
    elements = []

    if n > 0:
        fmt = ",".join(15 * ["5s"])
        line = f.next(skip_empty=True, comments="#").strip()
        data = read_record(line, fmt)
        elements += data

        while len(elements) < n:
            line = f.next().strip()
            data = read_record(line, fmt)
            elements += data

        return prune_values(elements[:n])

    elif n < 0:
        line = f.next(skip_empty=True, comments="#").strip()
        elements.append(line[:5].strip())
        while True:
            line = f.next().strip()

            if line:
                elements.append(line[:5].strip())

            else:
                break

        return elements

    else:
        _ = f.next(comments="#")  # Skip next blank line


def _read_indices_names(f, n):
    """Read indices or names."""
    if n > 0:
        line = f.next(comments="#").strip()
        data = [int(x) for x in line.split()]
        out = data

        return out[:n]

    elif n < 0:
        line = f.next(comments="#").strip()
        out = [line[:20].strip()]
        while True:
            line = f.next().strip()

            if line:
                out.append(line[:20].strip())

            else:
                break

        return out

    else:
        _ = f.next(comments="#")  # Skip next blank line


def _read_default(f, mopr_11=0):
    """Read default chemical property zones."""
    line = f.next(skip_empty=True, comments="#").strip()
    data = line.split()
    if len(data) < 9:
        raise ReadError()
    if mopr_11 == 1 and len(data) < 10:
        raise ReadError()

    default = {
        "default": {
            "initial_water": int(data[0]),
            "injection_water": int(data[1]),
            "mineral": int(data[2]),
            "initial_gas": int(data[3]),
            "adsorption": int(data[4]),
            "cation_exchange": int(data[5]),
            "permeability_porosity": int(data[6]),
            "linear_kd": int(data[7]),
            "injection_gas": int(data[8]),
        }
    }

    if len(data) == 10:
        if mopr_11 != 1:
            default["default"]["element"] = int(data[9])

        else:
            default["default"]["sedimentation_velocity"] = to_float(data[9])

    return default


def _read_zones(f, mopr_11=0):
    """Read chemical property zones."""
    zones = {"zones": {}}

    line = f.next(skip_empty=True, comments="#").rstrip()
    while True:
        # Normally, the list should end with a blank record
        # But some sample files don't yet end with an 'end' statement
        if not line.strip() or line.startswith("end"):
            break

        label = line[:5]
        data = line[5:].split()
        if len(data) < 11:
            raise ReadError()

        zones["zones"][label] = _parse_zones(data[2:], mopr_11)
        zones["zones"][label]["nseq"] = int(data[0])
        zones["zones"][label]["nadd"] = int(data[1])
        line = f.next(comments="#").rstrip()

    return zones


def _read_convergence_bounds(f):
    """Read convergence bounds."""
    while True:
        try:
            line = f.next().rstrip()

            if line.startswith("end"):
                return

            if line.startswith("CONVP"):
                break

        except StopIteration:
            return

    line = f.next(skip_empty=True, comments="#").strip()
    data = line.split()
    if len(data) < 4:
        raise ReadError()

    options = {
        "n_iteration_1": int(data[0]),
        "n_iteration_2": int(data[1]),
        "n_iteration_3": int(data[2]),
        "n_iteration_4": int(data[3]),
    }

    line = f.next(skip_empty=True, comments="#").strip()
    data = [to_float(x) for x in line.split()]
    if len(data) < 6:
        raise ReadError()

    options.update(
        {
            "t_increase_factor_1": data[0],
            "t_increase_factor_2": data[1],
            "t_increase_factor_3": data[2],
            "t_reduce_factor_1": data[3],
            "t_reduce_factor_2": data[4],
            "t_reduce_factor_3": data[5],
        }
    )

    return options


def _parse_zones(data, mopr_11):
    """Parse zones."""
    out = {
        "initial_water": int(data[0]),
        "injection_water": int(data[1]),
        "mineral": int(data[2]),
        "initial_gas": int(data[3]),
        "adsorption": int(data[4]),
        "cation_exchange": int(data[5]),
        "permeability_porosity": int(data[6]),
        "linear_kd": int(data[7]),
        "injection_gas": int(data[8]),
    }

    if len(data) == 10:
        if mopr_11 != 1:
            out["element"] = int(data[9])

        else:
            out["sedimentation_velocity"] = to_float(data[9])

    return out



from functools import wraps

import numpy as np

from ...._common import prune_values
from ..._common import read_record, write_record

str_to_dtype = {
    "int": (int, np.int32, np.int64),
    "float": (float, np.float32, np.float64),
    "str": (str,),
    "bool": (bool,),
    "str_int": (str, int, np.int32, np.int64),
    "array_like": (list, tuple, np.ndarray),
    "dict": (dict,),
    "scalar": (int, float, np.int32, np.int64, np.float32, np.float64),
    "scalar_array_like": (
        int,
        float,
        list,
        tuple,
        np.int32,
        np.int64,
        np.float32,
        np.float64,
        np.ndarray,
    ),
    "str_array_like": (str, list, tuple, np.ndarray),
}


def block(keyword, multi=False, noend=False):
    """Decorate block writing functions."""

    def decorator(func):
        from ._common import header

        @wraps(func)
        def wrapper(*args, **kwargs):
            head = f"{keyword:5}{header}"
            out = [head if noend else f"{head}\n"]
            out += func(*args, **kwargs)
            out += ["\n"] if multi else []

            return out

        return wrapper

    return decorator


def read_model_record(line, fmt, i=2):
    """Read model record defined by 'id' and 'parameters'."""
    data = read_record(line, fmt)

    return {
        "id": data[0],
        "parameters": prune_values(data[i:]),
    }


def write_model_record(data, key, fmt, space_between_values=False):
    """Write model record defined by 'id' and 'parameters'."""
    if key in data:
        values = [data[key]["id"], None]
        values += list(data[key]["parameters"])

        return write_record(values, fmt, space_between_values)

    else:
        return write_record([], [], space_between_values)


def read_primary_variables(f, fmt, n_variables):
    """Read primary variables."""
    data = []

    if n_variables:
        n = int(np.ceil(n_variables / 4))

        for _ in range(n):
            line = f.next()
            data += read_record(line, fmt)

    else:
        while True:
            i = f.tell()
            line = f.next()

            if line.strip():
                try:
                    data += read_record(line, fmt)

                except ValueError:
                    break

            else:
                break

        f.seek(i, increment=-1)

    return data



from . import _cli, capillarity, meshmaker, relative_permeability
from .__about__ import __version__
from ._helpers import convert_labels
from ._io import (
    Output,
    read_input,
    read_output,
    read_table,
    register_input,
    register_output,
    register_table,
    write_h5,
    write_input,
    write_output,
)
from ._mesh import CellBlock, Mesh, from_meshio, from_pyvista
from ._mesh import read as read_mesh
from ._mesh import read_time_series
from ._mesh import register as register_mesh
from ._mesh import write as write_mesh
from ._mesh import write_time_series
from ._run import run

__all__ = [
    "Mesh",
    "CellBlock",
    "Output",
    "meshmaker",
    "register_input",
    "register_output",
    "register_mesh",
    "register_table",
    "read_input",
    "read_output",
    "read_table",
    "write_h5",
    "write_input",
    "write_output",
    "from_meshio",
    "from_pyvista",
    "read_mesh",
    "write_mesh",
    "read_time_series",
    "write_time_series",
    "relative_permeability",
    "capillarity",
    "convert_labels",
    "run",
    "_cli",
    "__version__",
]



import numpy as np

from ...._common import open_file

__all__ = [
    "read",
]


def read(filename):
    """
    Read COLUMN table file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.

    Returns
    -------
    dict
        Table data.

    """
    with open_file(filename, "r") as f:
        line1 = f.readline().rstrip()
        line2 = f.readline().rstrip()

        # Find the number of columns
        line3 = f.readline().rstrip()
        tmp = line3

        ncol = 0
        ncols = [0]
        while tmp.strip():
            ncol = len(tmp) - len(tmp.lstrip())
            ncol += len(tmp.split()[0])
            tmp = tmp[ncol:]
            ncols.append(ncol)

        icols = np.cumsum(ncols)

        # Gather data
        data = [[float(x) for x in line3.split()]]

        for line in f:
            data += [[float(x) for x in line.split()]]

        data = np.transpose(data)

        # Parse headers
        headers = []
        for i1, i2 in zip(icols[:-1], icols[1:]):
            h1 = line1[i1:i2].strip()
            h2 = line2[i1:i2].strip()
            headers.append(f"{h1} {h2}")

        return {k: v for k, v in zip(headers, data)}



from functools import wraps


def section(header, footer="*"):
    """Decorate section writing functions."""

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            out = [f"{header}"]
            out += func(*args, **kwargs)
            out += [f"{footer}"]

            return out

        return wrapper

    return decorator



import numpy as np
import matplotlib.pyplot as plt


def permeability(porosity, phi_r=0.8, Gamma=0.8):
    theta = (porosity - phi_r) / (1.0 - phi_r)
    omega = 1.0 + (1.0 / Gamma) / (1.0 / phi_r - 1.0)

    return theta**2 * (1.0 - Gamma + Gamma / omega**2) / (1.0 - Gamma + Gamma * (theta / (theta + omega - 1.0))**2)


# Calculate porosity-permeability relationship
phi = np.linspace(0.80, 1.0, 201)
k = permeability(phi)

# Plot figure
style = {
    "font.size": 20,
    "xtick.labelsize": 20,
    "ytick.labelsize": 20,
}
plt.rcParams.update(style)

fig, ax = plt.subplots(1, 1, figsize=(8, 6))
ax.plot(phi, k, c="black", lw=2)
ax.set_xlabel("$\phi / \phi_0$")
ax.set_ylabel("$k / k_0$")
ax.set_xlim(phi.min(), phi.max())
ax.set_ylim(k.min(), k.max())
ax.set_xticks(np.linspace(0.8, 1.0, 5))

fig.tight_layout()
# fig.savefig("porosity_permeability.png", dpi=300)



import glob
import os
import pathlib
import platform
import shutil
import subprocess
import tempfile

_check_exec = True  # Bool to be monkeypatched in tests


def run(
    exec,
    input_filename,
    other_filenames=None,
    command=None,
    workers=None,
    docker=None,
    wsl=False,
    working_dir=None,
    use_temp=False,
    ignore_patterns=None,
    silent=False,
    petsc_args=None,
    docker_args=None,
    **kwargs,
):
    """
    Run TOUGH executable.

    Parameters
    ----------
    exec : str or pathlike
        Path to TOUGH executable.
    input_filename : str or pathlike
        TOUGH input file name.
    other_filenames : list, dict or None, optional, default None
        Other simulation files to copy to working directory (e.g., MESH, INCON, GENER) if not already present. If ``other_filenames`` is a dict, must be in the form ``{old: new}``, where ``old`` is the current name of the file to copy, and ``new`` is the name of the file copied.
    command : callable or None, optional, default None
        Command to execute TOUGH. Must be in the form ``f(exec, inp, [out])``, where ``exec`` is the path to TOUGH executable, ``inp`` is the input file name, and ``out`` is the output file name (optional).
    workers : int or None, optional, default None
        Number of MPI workers to invoke.
    docker : str, optional, default None
        Name of Docker image.
    wsl : bool, optional, default False
        Only for Windows. If `True`, run the final command as a Bash command.
    working_dir : str, pathlike or None, optional, default None
        Working directory. Input and output files will be generated in this directory.
    use_temp : bool, optional, default False
        If `True`, run simulation in a temporary directory, and copy simulation files to `working_dir` at the end of the simulation. This option may be required when running TOUGH through a Docker.
    ignore_patterns : list or None, optional, default None
        If provided, output files that match the glob-style patterns will be discarded.
    silent : bool, optional, default False
        If `True`, nothing will be printed to standard output.
    petsc_args : list or None, optional, default None
        List of arguments passed to PETSc solver (written to `.petscrc`).
    docker_args : list or None, optional, default None
        List of arguments passed to `docker run` command.

    Other Parameters
    ----------------
    block : str {'all', 'gener', 'mesh', 'incon'} or None, optional, default None
        Only if ``file_format = "tough"``. Blocks to be written:

         - 'all': write all blocks,
         - 'gener': only write block GENER,
         - 'mesh': only write blocks ELEME, COORD and CONNE,
         - 'incon': only write block INCON,
         - None: write all blocks except blocks defined in `ignore_blocks`.

    ignore_blocks : list of str or None, optional, default None
        Only if ``file_format = "tough"`` and `block` is None. Blocks to ignore.
    space_between_blocks : bool, optional, default False
        Only if ``file_format = "tough"``. Add an empty record between blocks.
    space_between_blocks : bool, optional, default True
        Only if ``file_format = "tough"``. Add a white space between floating point values.
    eos : str or None, optional, default None
        Only if ``file_format = "tough"``. Equation of State.
        If `eos` is defined in `parameters`, this option will be ignored.
    mopr_10 : int, optional, default 0
        Only if ``file_format = "toughreact-solute"``. MOPR(10) value in file 'flow.inp'.
    mopr_11 : int, optional, default 0
        Only if ``file_format = "toughreact-solute"``. MOPR(11) value in file 'flow.inp'.
    verbose : bool, optional, default True
        Only if ``file_format`` in {"toughreact-solute", "toughreact-chemical"}. If `True`, add comments to describe content of file.

    Returns
    -------
    :class:`subprocess.CompletedProcess`
        Subprocess completion status.

    """
    from . import write_input

    other_filenames = (
        {k: k for k in other_filenames}
        if isinstance(other_filenames, (list, tuple))
        else other_filenames
        if other_filenames
        else {}
    )

    if command is None:
        command = lambda exec, inp, out: f"{exec} {inp} {out}"

    ignore_patterns = list(ignore_patterns) if ignore_patterns else []
    ignore_patterns += [".OUTPUT*", "TABLE", "MESHA", "MESHB"]

    # Executable
    exec = str(exec)
    exec = f"{os.path.expanduser('~')}/{exec[2:]}" if exec.startswith("~/") else exec

    if _check_exec:
        if not docker:
            if shutil.which(exec) is None:
                raise RuntimeError(f"executable '{exec}' not found.")

        else:
            # Check if Docker is in the PATH
            if shutil.which("docker") is None:
                raise RuntimeError("Docker executable not found.")

            # Check if Docker daemon is running
            status = subprocess.run(
                ["docker", "version"],
                stdout=subprocess.PIPE,
                universal_newlines=True,
            )

            if "Server" not in status.stdout:
                raise RuntimeError(
                    "cannot connect to the Docker daemon. Is the Docker daemon running?"
                )

            # Check if Docker image exists
            status = subprocess.run(
                ["docker", "images", "-q", docker],
                stdout=subprocess.PIPE,
                universal_newlines=True,
            )

            if not status.stdout:
                raise RuntimeError(f"image '{docker}' not found.")

            # Check if the executable exists inside the image
            status = subprocess.run(
                ["docker", "run", "--rm", docker, "which", exec],
                stdout=subprocess.PIPE,
                universal_newlines=True,
            )

            if not status.stdout:
                raise RuntimeError(
                    f"executable '{exec}' not found in Docker image '{docker}'."
                )

    # Working directory
    working_dir = os.getcwd() if working_dir is None else working_dir
    working_dir = pathlib.Path(working_dir)
    working_dir.mkdir(parents=True, exist_ok=True)

    # Simulation directory
    if use_temp:
        temp_dir = tempfile.mkdtemp()
        simulation_dir = pathlib.Path(temp_dir)

        with open(working_dir / "tempdir.txt", "w") as f:
            f.write(temp_dir)

    else:
        simulation_dir = working_dir

    # Check if input file is in simulation directory, otherwise copy
    if not isinstance(input_filename, dict):
        input_path = pathlib.Path(input_filename)
        input_filename = simulation_dir / input_path.name

        if input_path.parent.resolve() != simulation_dir.resolve():
            shutil.copy(input_path, input_filename)

    else:
        write_input(simulation_dir / "INFILE", input_filename, **kwargs)
        input_filename = simulation_dir / "INFILE"

    # Copy other simulation files to working directory
    for k, v in other_filenames.items():
        filename = pathlib.Path(k)
        new_filename = pathlib.Path(v)

        if (
            filename.parent.resolve() != simulation_dir.resolve()
            or filename.name != new_filename.name
        ):
            shutil.copy(filename, simulation_dir / new_filename.name)

    # PETSc arguments
    petsc_args = petsc_args if petsc_args else []

    if petsc_args:
        with open(simulation_dir / ".petscrc", "w") as f:
            for arg in petsc_args:
                if arg.startswith("-"):
                    f.write(f"{arg} ")

                else:
                    f.write(f"{arg}\n")

    # Output filename
    output_filename = f"{input_filename.stem}.out"

    # TOUGH command
    cmd = command(exec, str(input_filename.name), str(output_filename))

    # Use MPI
    if workers is not None and workers > 1:
        cmd = f"mpiexec -n {workers} {cmd}"

    # Use Docker
    is_windows = platform.system().startswith("Win")

    if docker:
        if is_windows and os.getenv("ComSpec").endswith("cmd.exe"):
            cwd = '"%cd%"'

        else:
            cwd = "${PWD}"

        try:
            uid = f"-e LOCAL_USER_ID={os.getuid()}"

        except AttributeError:
            uid = ""

        docker_args = docker_args if docker_args else []
        docker_args += [
            "--rm",
            uid,
            "-v",
            f"{cwd}:/shared",
            "-w",
            "/shared",
        ]
        cmd = f"docker run {' '.join(docker_args)} {docker} {cmd}"

    # Use WSL
    if wsl and is_windows:
        cmd = f"bash -c '{cmd}'"

    # Run simulation
    if not silent:
        # See <https://www.koldfront.dk/making_subprocesspopen_in_python_3_play_nice_with_elaborate_output_1594>
        p = subprocess.Popen(
            cmd,
            shell=True,
            cwd=str(simulation_dir),
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            universal_newlines=False,
        )

        stdout = []
        cr = False
        for line in open(os.dup(p.stdout.fileno()), newline=""):
            # Handle carriage return
            # newline in open is converting \r\n as \r moving \r at the end of the previous string
            # This is only an issue for Spyder
            line = f"\r{line}" if cr else line
            cr = line.endswith("\r")

            line = line[:-2] if cr else line
            print(line, end="", flush=True)
            stdout.append(line)

        status = subprocess.CompletedProcess(
            args=p.args,
            returncode=0,
            stdout="".join(stdout),
        )

    else:
        status = subprocess.run(
            cmd,
            shell=True,
            cwd=str(simulation_dir),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
        )

    # Copy files from temporary directory and delete it
    if use_temp:
        shutil.copytree(
            simulation_dir,
            working_dir,
            ignore=shutil.ignore_patterns(*ignore_patterns),
            dirs_exist_ok=True,  # Doesn't work with Python 3.7
        )
        shutil.rmtree(simulation_dir, ignore_errors=True)
        os.remove(working_dir / "tempdir.txt")

    # Clean up working directory
    patterns = [
        pathlib.Path(filename)
        for pattern in ignore_patterns
        for filename in glob.glob(f"{str(simulation_dir)}/{pattern}")
    ]

    for pattern in patterns:
        os.remove(pattern)

    return status



import pathlib

import h5py

from ..output import Output
from ..output import read as read_output
from ..table import read as read_table


def write(
    filename,
    elements=None,
    connections=None,
    element_history=None,
    connection_history=None,
    generator_history=None,
    rock_history=None,
    labels_order=None,
    compression_opts=4,
):
    """
    Write TOUGH outputs to a HDF5 file.

    Parameters
    ----------
    filename : str or pathlike
        Output file name.
    elements : namedtuple, list of namedtuple, str, pathlike or None, optional, default None
        Element outputs to export.
    connections : namedtuple, list of namedtuple, str, pathlike or None, optional, default None
        Connection outputs to export.
    element_history : dict or None, optional, default None
        Element history to export.
    connection_history : dict or None, optional, default None
        Connection history to export.
    generator_history : dict or None, optional, default None
        Generator history to export.
    rock_history : dict or None, optional, default None
        Rock history to export.
    labels_order : list of array_like or None, optional, default None
        List of labels.
    compression_opts : int, optional, default 4
        Compression level for gzip compression. May be an integer from 0 to 9.

    """
    kwargs = {
        "compression": "gzip",
        "compression_opts": compression_opts,
    }

    with h5py.File(filename, "w") as f:
        if elements is not None:
            group = f.create_group("elements")
            _write_output(group, elements, labels_order, connection=False, **kwargs)

        if connections is not None:
            group = f.create_group("connections")
            _write_output(group, connections, labels_order, connection=True, **kwargs)

        if element_history is not None:
            group = f.create_group("element_history")
            _write_table(group, element_history, **kwargs)

        if connection_history is not None:
            group = f.create_group("connection_history")
            _write_table(group, connection_history, **kwargs)

        if generator_history is not None:
            group = f.create_group("generator_history")
            _write_table(group, generator_history, **kwargs)

        if rock_history is not None:
            group = f.create_group("rock_history")
            _write_table(group, rock_history, **kwargs)


def _write_output(f, outputs, labels_order, connection, **kwargs):
    """Write TOUGH output to group."""
    if isinstance(outputs, (str, pathlib.Path)):
        outputs = read_output(outputs, labels_order=labels_order, connection=connection)

    if isinstance(outputs, Output):
        outputs = [outputs]

    elif isinstance(outputs, (list, tuple)):
        for output in outputs:
            if not isinstance(output, Output):
                raise ValueError()

    else:
        raise ValueError()

    for output in outputs:
        group = f.create_group(f"time={output.time}")
        group.create_dataset("labels", data=output.labels.astype("S"), **kwargs)

        for k, v in output.data.items():
            group.create_dataset(k, data=v, **kwargs)


def _write_table(f, tables, **kwargs):
    """Write TOUGH table to group."""
    if not isinstance(tables, dict):
        raise ValueError()

    for name, table in tables.items():
        group = f.create_group(name)

        if isinstance(table, (str, pathlib.Path)):
            table = read_table(table)

        for k, v in table.items():
            group.create_dataset(k, data=v, **kwargs)



from ...._common import open_file
from .._common import getval, write_ffrecord

__all__ = [
    "write",
]


def write(filename, parameters, mopr_10=0, mopr_11=0, verbose=True):
    """
    Write TOUGHREACT solute input file.

    Parameters
    ----------
    filename : str
        Input file name.
    parameters : dict
        Parameters to export.
    mopr_10 : int, optional, default 0
        MOPR(10) value in file 'flow.inp'.
    mopr_11 : int, optional, default 0
        MOPR(11) value in file 'flow.inp'.
    verbose : bool, optional, default True
        If `True`, add comments to describe content of file.

    """
    buffer = write_buffer(parameters, mopr_10, mopr_11, verbose)
    with open_file(filename, "w") as f:
        for record in buffer:
            f.write(record)


def write_buffer(parameters, mopr_10=0, mopr_11=0, verbose=True, sections=None):
    """Write TOUGHREACT solute input file."""
    from ._common import sections as sections_

    # Section filters
    if sections is None:
        sections = sections_.copy()

    sections = set(sections)

    if sections.intersection(
        [
            "output",
            "elements",
            "components",
            "minerals",
            "aqueous_species",
            "surface_complexes",
            "exchange_species",
        ]
    ):
        numbers = _generate_numbers(parameters)

    # Define input file contents
    out = []

    if "title" in sections:
        out += _write_title(parameters, verbose)

    if "options" in sections:
        out += _write_options(parameters, verbose)

    if "filenames" in sections:
        out += _write_filenames(parameters, verbose)

    if "weights_coefficients" in sections:
        out += _write_weights_coefficients(parameters, verbose)

    if "convergence" in sections:
        out += _write_convergence(parameters, verbose)

    if "output" in sections:
        out += _write_output(parameters, numbers, verbose)

    if "elements" in sections:
        out += _write_elements(parameters, numbers["NWNOD"], verbose)

    if "components" in sections:
        out += _write_indices_names(parameters, "components", numbers["NWCOM"], verbose)

    if "minerals" in sections:
        out += _write_indices_names(parameters, "minerals", numbers["NWMIN"], verbose)

    if "aqueous_species" in sections:
        out += _write_indices_names(
            parameters, "aqueous_species", numbers["NWAQ"], verbose
        )

    if "surface_complexes" in sections:
        out += _write_indices_names(
            parameters, "surface_complexes", numbers["NWADS"], verbose
        )

    if "exchange_species" in sections:
        out += _write_indices_names(
            parameters, "exchange_species", numbers["NWEXC"], verbose
        )

    if "default" in sections:
        out += _write_default(parameters, verbose, mopr_11)

    if "zones" in sections:
        out += _write_zones(parameters, verbose, mopr_11)

    if "convergence_bounds" in sections:
        out += _write_convergence_bounds(parameters, verbose, mopr_10)

    if "end" in sections:
        out += ["end"]

    if "end_comments" in sections and "end_comments" in parameters:
        out += _write_end_comments(parameters)

    return "\n".join(out)


def _generate_numbers(parameters):
    """Generate numbers of outputs."""
    output = parameters["output"]
    nwnod = len(output["elements"]) if "elements" in output else 0
    nwcom = len(output["components"]) if "components" in output else 0
    nwmin = len(output["minerals"]) if "minerals" in output else 0
    nwaq = len(output["aqueous_species"]) if "aqueous_species" in output else 0
    nwads = len(output["surface_complexes"]) if "surface_complexes" in output else 0
    nwexc = len(output["exchange_species"]) if "exchange_species" in output else 0

    numbers = {
        "NWNOD": -nwnod,
        "NWCOM": nwcom,
        "NWMIN": nwmin,
        "NWAQ": nwaq,
        "NWADS": nwads,
        "NWEXC": nwexc,
    }

    if numbers["NWCOM"]:
        if isinstance(parameters["output"]["components"][0], str):
            numbers["NWCOM"] *= -1

    if numbers["NWMIN"]:
        if isinstance(parameters["output"]["minerals"][0], str):
            numbers["NWMIN"] *= -1

    if numbers["NWAQ"]:
        if isinstance(parameters["output"]["aqueous_species"][0], str):
            numbers["NWAQ"] *= -1

    if numbers["NWADS"]:
        if isinstance(parameters["output"]["surface_complexes"][0], str):
            numbers["NWADS"] *= -1

    if numbers["NWEXC"]:
        if isinstance(parameters["output"]["exchange_species"][0], str):
            numbers["NWEXC"] *= -1

    return numbers


def _write_title(parameters, verbose):
    """Write title."""
    out = ["# Title"] if verbose else []
    out += [getval(parameters, "title", "")]

    return out


def _write_options(parameters, verbose):
    """Write options."""
    out = []

    # Record 2
    values = [
        to_int(getval(parameters, ("flags", "iteration_scheme"), 0)),
        to_int(getval(parameters, ("flags", "reactive_surface_area"), 0)),
        to_int(getval(parameters, ("flags", "solver"), 0)),
        to_int(getval(parameters, ("flags", "n_subiteration"), 0)),
        to_int(getval(parameters, ("flags", "gas_transport"), 0)),
        to_int(getval(parameters, ("flags", "verbosity"), 0)),
        to_int(getval(parameters, ("flags", "feedback"), 0)),
        to_int(getval(parameters, ("flags", "coupling"), 0)),
        0,
    ]

    out += (
        ["# ISPIA ITERSFA  ISOLVC   NGAMM   NGAS1 ICHDUMP    KCPL ICO2H2O iTDS_REACT"]
        if verbose
        else []
    )
    fmt = ["{:7d}"] * 8 + ["{:10d}"]
    out += write_ffrecord(values, verbose, fmt)

    # Record 3
    values = [
        to_float(getval(parameters, ("options", "sl_min"), 0.0)),
        to_float(getval(parameters, ("options", "rcour"), 0.0)),
        to_float(getval(parameters, ("options", "ionic_strength_max"), 0.0)),
        to_float(getval(parameters, ("options", "mineral_gas_factor"), 0.0)),
    ]

    out += ["#  SL1MIN     RCOUR    STIMAX    CNFACT"] if verbose else []
    out += write_ffrecord(values, verbose, float_fmt="{{:9f}}")

    return out


def _write_filenames(parameters, verbose):
    """Write file names."""
    out = ["# Input and output file names"] if verbose else []
    values = [
        str(getval(parameters, ("files", "thermodynamic_input"), "")),
        str(getval(parameters, ("files", "iteration_output"), "")),
        str(getval(parameters, ("files", "plot_output"), "")),
        str(getval(parameters, ("files", "solid_output"), "")),
        str(getval(parameters, ("files", "gas_output"), "")),
        str(getval(parameters, ("files", "time_output"), "")),
    ]

    if verbose:
        out += [f"{values[0]:<20}  ! Thermodynamic database"]
        out += [f"{values[1]:<20}  ! Iteration information"]
        out += [f"{values[2]:<20}  ! Aqueous concentrations in Tecplot form"]
        out += [f"{values[3]:<20}  ! Mineral data in Tecplot form"]
        out += [f"{values[4]:<20}  ! Gas data in Tecplot form"]
        out += [f"{values[5]:<20}  ! Concentrations at specific elements over time"]

    else:
        out += [f"{value}" for value in values]

    return out


def _write_weights_coefficients(parameters, verbose):
    """Write weights and diffusion coefficients."""
    out = ["#   ITIME      WUPC     DFFUN    DFFUNG"] if verbose else []
    values = [
        to_float(getval(parameters, ("options", "w_time"), 0.0)),
        to_float(getval(parameters, ("options", "w_upstream"), 0.0)),
        to_float(getval(parameters, ("options", "aqueous_diffusion_coefficient"), 0.0)),
        to_float(
            getval(parameters, ("options", "molecular_diffusion_coefficient"), 0.0)
        ),
    ]

    out += write_ffrecord(values, verbose, float_fmt="{{:9f}}")

    return out


def _write_convergence(parameters, verbose):
    """Write convergence criterion."""
    out = (
        ["#MAXITPTR     TOLTR  MAXITPCH     TOLCH     TOLMB     TOLDC     TOLDR"]
        if verbose
        else []
    )
    values = [
        to_int(getval(parameters, ("options", "n_iteration_tr"), 0)),
        to_float(getval(parameters, ("options", "eps_tr"), 0.0)),
        to_int(getval(parameters, ("options", "n_iteration_ch"), 0)),
        to_float(getval(parameters, ("options", "eps_ch"), 0.0)),
        to_float(getval(parameters, ("options", "eps_mb"), 0.0)),
        to_float(getval(parameters, ("options", "eps_dc"), 0.0)),
        to_float(getval(parameters, ("options", "eps_dr"), 0.0)),
    ]

    out += write_ffrecord(values, verbose, int_fmt="{:9d}", float_fmt="{{:9f}}")

    return out


def _write_output(parameters, numbers, verbose):
    """Write output control variables."""
    out = (
        ["# NWTI  NWNOD  NWCOM  NWMIN   NWAQ  NWADS  NWEXC   ICON    MIN   IGAS"]
        if verbose
        else []
    )
    values = [
        to_int(getval(parameters, ("options", "n_cycle_print"), 0)),
        to_int(numbers["NWNOD"]),
        to_int(numbers["NWCOM"]),
        to_int(numbers["NWMIN"]),
        to_int(numbers["NWAQ"]),
        to_int(numbers["NWADS"]),
        to_int(numbers["NWEXC"]),
        to_int(getval(parameters, ("flags", "aqueous_concentration_unit"), 0)),
        to_int(getval(parameters, ("flags", "mineral_unit"), 0)),
        to_int(getval(parameters, ("flags", "gas_concentration_unit"), 0)),
    ]

    out += write_ffrecord(values, verbose, int_fmt="{:6d}")

    return out


def _write_elements(parameters, n, verbose):
    """Write list of grid blocks."""
    out = [f"# {comments['nodes']}"] if verbose else []

    if n == 0:
        out += [""]

        return out

    out += [f"{x[:5]}" for x in getval(parameters, ("output", "elements"), [])]
    out += [""]

    return out


def _write_indices_names(parameters, key, n, verbose):
    """Write indices or names."""
    out = [f"# {comments[key]}"] if verbose else []

    if n == 0:
        out += [""]

        return out

    if n > 0:
        out += [f"{' '.join(str(x) for x in getval(parameters, ('output', key), []))}"]

    else:
        out += [f"{x[:20]}" for x in getval(parameters, ("output", key), [])]
        out += [""]

    return out


def _write_default(parameters, verbose, mopr_11=0):
    """Write default chemical property zones."""
    out = (
        ["#IZIWDF IZBWDF IZMIDF IZGSDF IZADDF IZEXDF IZPPDF IZKDDF IZBGDF"]
        if verbose
        else []
    )
    values = [
        to_int(getval(parameters, ("default", "initial_water"), 0)),
        to_int(getval(parameters, ("default", "injection_water"), 0)),
        to_int(getval(parameters, ("default", "mineral"), 0)),
        to_int(getval(parameters, ("default", "initial_gas"), 0)),
        to_int(getval(parameters, ("default", "adsorption"), 0)),
        to_int(getval(parameters, ("default", "cation_exchange"), 0)),
        to_int(getval(parameters, ("default", "permeability_porosity"), 0)),
        to_int(getval(parameters, ("default", "linear_kd"), 0)),
        to_int(getval(parameters, ("default", "injection_gas"), 0)),
    ]

    if mopr_11 != 1:
        if "element" in parameters["default"]:
            values += [parameters["default"]["element"]]

    else:
        values += [getval(parameters, ("default", "sedimentation_velocity"), 0)]

    tmp = write_ffrecord(values, verbose, int_fmt="{:6d}", float_fmt="{{:9f}}")
    out += [f" {tmp[0]}"] if verbose else tmp

    return out


def _write_zones(parameters, verbose, mopr_11=0):
    """Write chemical property zones."""
    out = (
        ["#ELEM NSEQ NADD IZIW IZBW IZMI IZGS IZAD IZEX IZPP IZKD IZBG"]
        if verbose
        else []
    )

    if "zones" not in parameters or not parameters["zones"]:
        out += [""]

        return out

    # Do not use items() for better warning logging in function getval
    for zone in parameters["zones"]:
        values = [
            zone,
            to_int(getval(parameters, ("zones", zone, "nseq"), 0)),
            to_int(getval(parameters, ("zones", zone, "nadd"), 0)),
            to_int(getval(parameters, ("zones", zone, "initial_water"), 0)),
            to_int(getval(parameters, ("zones", zone, "injection_water"), 0)),
            to_int(getval(parameters, ("zones", zone, "mineral"), 0)),
            to_int(getval(parameters, ("zones", zone, "initial_gas"), 0)),
            to_int(getval(parameters, ("zones", zone, "adsorption"), 0)),
            to_int(getval(parameters, ("zones", zone, "cation_exchange"), 0)),
            to_int(getval(parameters, ("zones", zone, "permeability_porosity"), 0)),
            to_int(getval(parameters, ("zones", zone, "linear_kd"), 0)),
            to_int(getval(parameters, ("zones", zone, "injection_gas"), 0)),
        ]

        if mopr_11 != 1:
            if "element" in parameters["zones"][zone]:
                values += [parameters["zones"][zone]["element"]]

        else:
            values += [
                getval(parameters, ("zones", zone, "sedimentation_velocity"), 0.0)
            ]

        out += write_ffrecord(
            values, verbose, int_fmt="{:4d}", float_fmt="{{:9f}}", str_fmt="{:5}"
        )

    out += [""]

    return out


def _write_convergence_bounds(parameters, verbose, mopr_10=0):
    """Write convergence bounds."""
    if mopr_10 != 2:
        return []

    out = ["CONVP"]
    out += ["# Read in chemical convergence limits if MOPR(10)=2"] if verbose else []

    # Numbers of iterations
    values = [
        to_int(getval(parameters, ("options", "n_iteration_1"), 30), 30),
        to_int(getval(parameters, ("options", "n_iteration_2"), 50), 50),
        to_int(getval(parameters, ("options", "n_iteration_3"), 75), 75),
        to_int(getval(parameters, ("options", "n_iteration_4"), 100), 100),
    ]

    out += ["#MAXCHEM1  MAXCHEM2  MAXCHEM3  MAXCHEM4"] if verbose else []
    out += write_ffrecord(values, verbose, int_fmt="{:9d}")

    # Increase/reduce factors
    values = [
        to_float(getval(parameters, ("options", "t_increase_factor_1"), 2.0), 2.0),
        to_float(getval(parameters, ("options", "t_increase_factor_2"), 1.5), 1.5),
        to_float(getval(parameters, ("options", "t_increase_factor_3"), 1.0), 1.0),
        to_float(getval(parameters, ("options", "t_reduce_factor_1"), 0.8), 0.8),
        to_float(getval(parameters, ("options", "t_reduce_factor_2"), 0.6), 0.6),
        to_float(getval(parameters, ("options", "t_reduce_factor_3"), 0.5), 0.5),
    ]

    out += (
        ["# DTINCR1   DTINCR2   DTINCR3   DTDECR1   DTDECR2   DTDECR3"]
        if verbose
        else []
    )
    out += write_ffrecord(values, verbose, float_fmt="{{:9f}}")

    return out


def _write_end_comments(parameters):
    """Write end comments."""
    end_comments = getval(parameters, "end_comments", "")

    return [end_comments] if isinstance(end_comments, str) else end_comments


def to_int(x, default=0):
    """Handle nones in integer conversion."""
    return int(x) if x is not None else default


def to_float(x, default=0.0):
    """Handle nones in floating point conversion."""
    return float(x) if x is not None else default


comments = {
    "nodes": "Nodes",
    "components": "Primary aqueous species",
    "minerals": "Minerals",
    "aqueous_species": "Individual aqueous species",
    "surface_complexes": "Adsorption species",
    "exchange_species": "Exchange species",
}



from .. import tough


def read(filename, eos=None):
    """
    Read TOUGHREACT flow input file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.
    eos : str or None, optional, default None
        Equation of State.

    Returns
    -------
    dict
        TOUGHREACT flow input parameters.

    """
    return tough.read(filename, label_length=5, eos=eos, simulator="toughreact")


def write(filename, parameters, block=None, ignore_blocks=None, eos=None):
    """
    Write TOUGHREACT flow input file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Output file name or buffer.
    parameters : dict
        Parameters to export.
    block : str {'all', 'gener', 'mesh', 'incon'} or None, optional, default None
        Blocks to be written:

         - 'all': write all blocks,
         - 'gener': only write block GENER,
         - 'mesh': only write blocks ELEME, COORD and CONNE,
         - 'incon': only write block INCON,
         - None: write all blocks except blocks defined in `ignore_blocks`.

    ignore_blocks : list of str or None, optional, default None
        Blocks to ignore. Only if `block` is None.
    eos : str or None, optional, default None
        Equation of State. If `eos` is defined in `parameters`, this option will be ignored.

    """
    return tough.write(
        filename, parameters, block, ignore_blocks, eos, simulator="toughreact"
    )



import pathlib

import h5py
import helpers

import toughio


def test_h5():
    this_dir = pathlib.Path(__file__).parent
    path = this_dir / "support_files" / "outputs"
    filename = helpers.tempdir("output.h5")

    toughio.write_h5(
        filename=filename,
        elements=path / "OUTPUT_ELEME.csv",
        connections=path / "OUTPUT_CONNE.csv",
        element_history={"A1912": path / "FOFT_A1912.csv"},
        connection_history={"A1912": path / "FOFT_A1912.csv"},
        generator_history={"A1162": path / "GOFT_A1162.csv"},
        rock_history={"ROCK": path / "ROFT.csv"},
    )

    with h5py.File(filename, "r") as f:
        assert list(f.keys()) == [
            "connection_history",
            "connections",
            "element_history",
            "elements",
            "generator_history",
            "rock_history",
        ]



"""
Generate MESH and INCON files
=============================

In this example, we assume that the mesh has already been generated by a third-party software (here Gmsh via :mod:`pygmsh`).

"""

########################################################################################
# First, we import :mod:`numpy` and :mod:`toughio`.

import numpy as np
import toughio

########################################################################################

########################################################################################
# A supported mesh can be read using the function :func:`toughio.read_mesh` that returns a :class:`toughio.Mesh` object.

mesh = toughio.read_mesh("mesh.vtu")

########################################################################################

########################################################################################
# (Optional) Although physical group names have been defined previously, most meshes do not export this information, including VTU.
# We can use the method :meth:`toughio.Mesh.add_material` to manually add material names.

mesh.add_material("UPPAQ", 1)
mesh.add_material("CAPRO", 2)
mesh.add_material("CENAQ", 3)
mesh.add_material("BASAQ", 4)
mesh.add_material("FAULT", 5)
mesh.add_material("BOUND", 6)

########################################################################################

########################################################################################
# The mesh used in this sample problem is 2D and has been defined in the XY plane, but the points have 3D coordinates (with zeros as 3rd dimension for every cells). To make it 3D in the XZ plane, we swap the 2nd and 3rd dimensions, and then extrude the mesh by 1 meter along the Y axis (2nd dimension).

mesh.points[:, [1, 2]] = mesh.points[:, [2, 1]]
mesh.extrude_to_3d(height=1.0, axis=1)

########################################################################################

########################################################################################
# (Optional) Before going any further, it is good practice to first check the quality of the mesh generated.
# TOUGH does not use any geometrical coordinate system and assumes that the line connecting a cell with its neighbor is orthogonal to their common interface. :mod:`toughio` provides a mesh property that measures the quality of a cell as the average absolute cosine angle between the line connecting a cell with its neighbor and the normal vector of the common interface.
# The mesh used in this example is of rather good quality. Bad quality cells are located at the boundaries of the model and mostly belong to the material `"BOUND"`. As this material is only used to impose Dirichlet boundary conditions in TOUGH, these bad quality cells will not impact the simulation outputs.

import pyvista
pyvista.set_plot_theme("document")

p = pyvista.Plotter(window_size=(800, 800))
p.add_mesh(
    mesh=mesh.to_pyvista(),
    scalars=mesh.qualities,
    scalar_bar_args={"title": "Average cell quality"},
    clim=(0.0, 1.0),
    cmap="RdBu",
    show_scalar_bar=True,
    show_edges=True,
)
p.view_xz()
p.show()

########################################################################################

########################################################################################
# (Optional) Usually, a simple distribution plot is enough to rapidly assess the quality of a mesh.

import seaborn

ax = seaborn.displot(mesh.qualities[mesh.materials != "BOUND"], kind="hist")

########################################################################################

########################################################################################
# We start by defining the boundary conditions. :mod:`toughio` recognizes the cell data key `"boundary_condition"` and automatically imposes Dirichlet boundary conditions to cells that have any value other than 0 in this cell data array.
# In this example, we simply set 1 to cells that belong to the group `"BOUND"` and 0 to others.

materials = mesh.materials
bcond = (materials == "BOUND").astype(int)
mesh.add_cell_data("boundary_condition", bcond)

########################################################################################

########################################################################################
# Initial conditions can be defined as a cell data array associated to key `"initial_condition"` where each column of the array corresponds to a primary variable. Note that :mod:`toughio` will not write any initial condition value that is lower than the threshold flag -1.0e9.

centers = mesh.centers
incon = np.full((mesh.n_cells, 4), -1.0e9)
incon[:, 0] = 1.0e5 - 9810.0 * centers[:, 2]
incon[:, 1] = 0.05
incon[:, 2] = 0.0
incon[:, 3] = 10.0 - 0.025 * centers[:, 2]
mesh.add_cell_data("initial_condition", incon)

########################################################################################

########################################################################################
# :mod:`toughio` also recognizes the cell data keys `"porosity"` and `"permeability"` in case we want to initialize porosity and/or permeability fields (e.g., if well logs data are available). Like boundary and initial conditions, we only have to associate new cell data arrays to keys `"porosity"` and/or `"permeability"`. The way these arrays are generated does not matter, they can be the results of simple interpolations (e.g., with :mod:`scipy`) or more advanced geostatistical interpolations (e.g., with :mod:`pykrige`).

########################################################################################

########################################################################################
# We can now write the `MESH` and `INCON` files by calling the method :meth:`toughio.Mesh.write_tough`.
# Additionally, we can also pickle the final mesh for later use (reading a pickle file is much faster than reading any mesh format).

mesh.write_tough("MESH", incon=True)
mesh.write("mesh.pickle")

########################################################################################



from ...._common import open_file
from ...._exceptions import ReadError
from ...._helpers import FileIterator
from ..._common import to_float
from .._common import read_end_comments

__all__ = [
    "read",
]


def read(filename):
    """
    Read TOUGHREACT chemical input file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.

    """
    with open_file(filename, "r") as f:
        out = read_buffer(f)

    return out


def read_buffer(f):
    """Read TOUGHREACT chemical input file."""
    parameters = {}
    fiter = FileIterator(f)

    try:
        parameters.update(_read_title(fiter))
        parameters.update(_read_prim(fiter))
        parameters.update(_read_akin(fiter))
        parameters.update(_read_aque(fiter))
        parameters.update(_read_miner(fiter))
        parameters.update(_read_gas(fiter))
        parameters.update(_read_surx(fiter))
        parameters.update(_read_kdde(fiter))

        exch, nxsites = _read_exch(fiter)
        parameters.update(exch)

        parameters["zones"] = {}
        parameters["zones"].update(_read_water(fiter))
        parameters["zones"].update(_read_imin(fiter))
        parameters["zones"].update(_read_igas(fiter))
        parameters["zones"].update(_read_zppr(fiter))
        parameters["zones"].update(_read_zads(fiter))
        parameters["zones"].update(_read_zlkd(fiter))
        parameters["zones"].update(_read_zexc(fiter, nxsites))

        # End comments
        end_comments = read_end_comments(fiter)
        if end_comments:
            parameters["end_comments"] = end_comments

    except:
        raise ReadError(f"failed to parse line {fiter.count}.")

    return parameters


def _read_title(f):
    """Read title."""
    line = f.next(skip_empty=True, comments="#").strip()

    return {"title": line}


def _read_prim(f):
    """Read primary species."""
    prim = {"primary_species": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    # Loop until reading character *
    while not line.startswith("*"):
        data = line.split()
        if len(data) < 2:
            raise ReadError()

        tmp = {"name": data[0], "transport": int(data[1])}
        if len(data) >= 3:
            tmp["mineral_name"] = data[2]
        if len(data) >= 4:
            tmp["sorption_density"] = to_float(data[3])
        if len(data) >= 5:
            tmp["adsorption_id"] = int(data[4])
        if len(data) >= 6:
            tmp["capacitance"] = to_float(data[5])
        prim["primary_species"].append(tmp)

        line = _nextline(f).strip()

    return prim


def _read_akin(f):
    """Read aqueous kinetics."""
    akin = {"aqueous_kinetics": None}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Record 2
    ntrx = int(line.strip())
    akin["aqueous_kinetics"] = []

    # Loop ntrx times
    for _ in range(ntrx):
        # Record 3
        line = _nextline(f).strip()
        # irx = int(line.strip())

        # Record 4
        line = _nextline(f).strip()
        data = line.split()
        ncp = int(data[0])
        if len(data) < 2 * ncp + 1:
            raise ReadError()

        names = data[2 : 2 * ncp + 2 : 2]
        coeffs = data[1 : 2 * ncp + 1 : 2]
        tmp = {
            "species": [
                {
                    "name": name,
                    "stoichiometric_coeff": to_float(coeff),
                }
                for name, coeff in zip(names, coeffs)
            ]
        }

        if len(data) > 2 * ncp + 1:
            tmp["reaction_affinity"] = {
                "id": int(data[2 * ncp + 1]),
                "cf": to_float(data[2 * ncp + 2]),
                "logK": to_float(data[2 * ncp + 3]),
            }

        # Record 5
        line = _nextline(f).strip()
        data = line.split()
        tmp["id"] = int(data[0])
        tmp["n_mechanism"] = int(data[1])

        # Record 6
        line = _nextline(f).strip()
        data = to_float(line.split()[0])

        if data != -1.0:
            tmp["rate"] = data

        else:
            # Record 6.1
            line = _nextline(f).strip()
            data = line.split()
            tmp["rate"] = {
                "k25": to_float(data[0]),
                "Ea": to_float(data[1]),
            }

        # Record 7
        line = _nextline(f).strip()
        data = line.split()
        ncp_rx1 = int(data[0])
        if len(data) < 3 * ncp_rx1 + 1:
            raise ReadError()

        tmp["product"] = [
            {
                "specie": data[3 * i + 1],
                "flag": int(data[3 * i + 2]),
                "power": to_float(data[3 * i + 3]),
            }
            for i in range(ncp_rx1)
        ]

        # Record 8
        line = _nextline(f).strip()
        data = line.split()
        ncp_rx2 = int(data[0])
        if len(data) < 3 * ncp_rx2 + 1:
            raise ReadError()

        tmp["monod"] = [
            {
                "specie": data[3 * i + 1],
                "flag": int(data[3 * i + 2]),
                "half_saturation": to_float(data[3 * i + 3]),
            }
            for i in range(ncp_rx2)
        ]

        # Record 9
        line = _nextline(f).strip()
        data = line.split()
        ncp_rx3 = int(data[0])
        if len(data) < 3 * ncp_rx3 + 1:
            raise ReadError()

        tmp["inhibition"] = [
            {
                "specie": data[3 * i + 1],
                "flag": int(data[3 * i + 2]),
                "constant": to_float(data[3 * i + 3]),
            }
            for i in range(ncp_rx3)
        ]

        akin["aqueous_kinetics"].append(tmp)

    # '*' is not used here to mark the end of the list. Skip it.
    _ = f.next()

    return akin


def _read_aque(f):
    """Read aqueous species."""
    aque = {"aqueous_species": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Loop until reading character *
    while not line.startswith("*"):
        data = line.strip()
        aque["aqueous_species"].append(data)

        line = _nextline(f)

    return aque


def _read_miner(f):
    """Read minerals."""
    miner = {"minerals": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Loop until reading character *
    while not line.startswith("*"):
        # Record 2
        data = _nextsplitline(line, 5)

        tmp = {
            "name": data[0],
            "type": int(data[1]),
            "kinetic_constraint": int(data[2]),
            "solid_solution": int(data[3]),
            "precipitation_dry": int(data[4]),
        }

        # Records 2.1 and 2.2
        # The 8 first entries of record 2.2 are common with record 2.1
        if tmp["type"] == 1:
            n = 2 if tmp["kinetic_constraint"] == 3 else 1
            for ir in range(n):
                if tmp["kinetic_constraint"] == 1:
                    key = "dissolution"

                elif tmp["kinetic_constraint"] == 2:
                    key = "precipitation"

                elif tmp["kinetic_constraint"] == 3:
                    key = "dissolution" if ir == 0 else "precipitation"

                data = (
                    _nextsplitline(f, 8)
                    if key == "dissolution"
                    else _nextsplitline(f, 10)
                )

                tmp[key] = {}
                tmp[key]["k25"] = to_float(data[0])
                tmp[key]["rate_ph_dependence"] = int(data[1])
                tmp[key]["eta"] = to_float(data[2])
                tmp[key]["theta"] = to_float(data[3])
                tmp[key]["activation_energy"] = to_float(data[4])
                tmp[key]["a"] = to_float(data[5])
                tmp[key]["b"] = to_float(data[6])
                tmp[key]["c"] = to_float(data[7])

                # Record 2.2
                if key == "precipitation":
                    tmp[key]["volume_fraction_ini"] = to_float(data[8])
                    tmp[key]["id"] = int(data[9])

                # Record 2.1.1
                if tmp[key]["rate_ph_dependence"] == 1:
                    data = _nextsplitline(f, 4)

                    tmp[key]["ph1"] = to_float(data[0])
                    tmp[key]["slope1"] = to_float(data[1])
                    tmp[key]["ph2"] = to_float(data[2])
                    tmp[key]["slope2"] = to_float(data[3])

                # Record 2.1.2
                elif tmp[key]["rate_ph_dependence"] == 2:
                    tmp[key]["extra_mechanisms"] = []

                    line = _nextline(f).strip()
                    ndis = int(line.strip())

                    # Record 2.1.2.1
                    for _ in range(ndis):
                        line = _nextline(f).strip()
                        data = line.split()
                        nspds = int(data[2])
                        if len(data) < 2 * nspds + 3:
                            raise ReadError()

                        tmp[key]["extra_mechanisms"].append(
                            {
                                "ki": to_float(data[0]),
                                "activation_energy": to_float(data[1]),
                                "species": [
                                    {
                                        "name": data[2 * i + 3],
                                        "power": to_float(data[2 * i + 4]),
                                    }
                                    for i in range(nspds)
                                ],
                            }
                        )

        # Record 3
        if tmp["kinetic_constraint"] != 1:
            data = _nextsplitline(f, 3)
            tmp["gap"] = to_float(data[0])
            tmp["temp1"] = to_float(data[1])
            tmp["temp2"] = to_float(data[2])

        miner["minerals"].append(tmp)
        line = _nextline(f)

    return miner


def _read_gas(f):
    """Read gaseous species."""
    gas = {"gaseous_species": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Loop until reading character *
    while not line.startswith("*"):
        data = line.split()
        if len(data) < 2:
            raise ReadError()

        tmp = {
            "name": data[0],
            "fugacity": int(data[1]),
        }
        gas["gaseous_species"].append(tmp)

        line = _nextline(f).strip()

    return gas


def _read_surx(f):
    """Read surface complexe."""
    surx = {"surface_complexes": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Loop until reading character *
    while not line.startswith("*"):
        data = line.strip()
        surx["surface_complexes"].append(data)

        line = _nextline(f)

    return surx


def _read_kdde(f):
    """Read Kd and decay."""
    kdde = {"kd_decay": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Loop until reading character *
    while not line.startswith("*"):
        data = line.split()
        if len(data) < 4:
            raise ReadError()

        tmp = {
            "name": data[0],
            "decay_constant": to_float(data[1]),
            "a": to_float(data[2]),
            "b": to_float(data[3]),
        }
        kdde["kd_decay"].append(tmp)

        line = _nextline(f).strip()

    return kdde


def _read_exch(f):
    """Read exchanged species."""
    exch = {"exchanged_species": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}, 0

    # Record 2
    data = line.split()
    if len(data) < 2:
        raise ReadError()

    nxsites = int(data[0])
    exch["exchange_sites_id"] = int(data[1])

    # Record 4
    line = _nextline(f, skip_empty=True, comments="#")

    # Loop until reading character *
    while not line.startswith("*"):
        data = line.split()
        if len(data) < nxsites + 3:
            raise ReadError()

        tmp = {
            "name": data[0],
            "reference": bool(int(data[1])),
            "type": int(data[2]),
            "site_coeffs": [to_float(x) for x in data[3:]],
        }
        exch["exchanged_species"].append(tmp)

        line = _nextline(f).strip()

    return exch, nxsites


def _read_water(f):
    """Read water zones."""
    zones = {"initial_waters": [], "injection_waters": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Record 2
    data = line.split()
    if len(data) < 2:
        raise ReadError()

    niwtype = int(data[0])
    nbwtype = int(data[1])

    # Loop niwtype + nbwtype times
    for i in range(niwtype + nbwtype):
        key = "initial_waters" if i < niwtype else "injection_waters"

        # Record 4
        line = _nextline(f, skip_empty=True, comments="#")
        data = line.split()
        iwtype = int(data[0])

        if iwtype >= 0 and len(data) < 3:
            raise ReadError()

        elif iwtype < 0 and len(data) < 4:
            raise ReadError()

        tmp = {
            "temperature": to_float(data[1]),
            "pressure": to_float(data[2]),
        }
        if iwtype < 0:
            tmp["rock"] = data[3]
        tmp["species"] = []

        # Record 6
        line = _nextline(f, remove_quotes=False, skip_empty=True, comments="#")
        while not line.replace("'", "").startswith("*"):
            data = _nextsplitline(line, 6)

            # split() doesn't work in this case as 'nameq' can be ' ' if not needed
            if data[4].startswith("'") and data[5].endswith("'"):
                nameq = (data[4] + data[5]).replace("'", "").strip()
                log_fugacity = data[6]

            else:
                nameq = data[4].replace("'", "")
                log_fugacity = data[5]

            tmp2 = {
                "name": data[0].replace("'", ""),
                "flag": int(data[1]),
                "guess": to_float(data[2]),
                "ctot": to_float(data[3]),
                "log_fugacity": to_float(log_fugacity),
            }
            if nameq:
                tmp2["nameq"] = nameq

            tmp["species"].append(tmp2)
            line = _nextline(f, remove_quotes=False).strip()

        zones[key].append(tmp)

    return zones


def _read_imin(f):
    """Read mineral zones."""
    zones = {"minerals": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Record 3
    nmtype = int(line.strip())

    # Loop nmtype times
    for _ in range(nmtype):
        # Record 4
        line = _nextline(f, skip_empty=True, comments="#")
        data = line.split()
        imtype = int(data[0])

        if imtype >= 0 and len(data) < 1:
            raise ReadError()

        elif imtype < 0 and len(data) < 2:
            raise ReadError()

        tmp = {}
        if imtype < 0:
            tmp["rock"] = data[1]
        tmp["species"] = []

        # Record 6
        line = _nextline(f, skip_empty=True, comments="#")
        while not line.startswith("*"):
            data = _nextsplitline(line, 3)
            tmp2 = {
                "name": data[0],
                "volume_fraction_ini": to_float(data[1]),
                "flag": int(data[2]),
            }

            # Record 6.1
            if tmp2["flag"] == 1:
                data = _nextsplitline(f, 3)

                tmp2["radius"] = to_float(data[0])
                tmp2["area_ini"] = to_float(data[1])
                tmp2["area_unit"] = int(data[2])

                # TODO: investigate what is the next record when IMFLG2 < 0
                # This record is not documented in the user guide (p. 65)
                if tmp2["area_unit"] < 0:
                    data = _nextsplitline(f, 4)
                    tmp2["unknown"] = [
                        int(data[0]),
                        to_float(data[1]),
                        int(data[2]),
                        to_float(data[3]),
                    ]

            tmp["species"].append(tmp2)
            line = _nextline(f).strip()

        zones["minerals"].append(tmp)

    return zones


def _read_igas(f):
    """Read gas zones."""
    zones = {"initial_gases": [], "injection_gases": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Record 3
    data = _nextsplitline(line, 2)
    nigtype = int(data[0])
    nbgtype = int(data[1])

    # Loop nigtype + nbgtype times
    for i in range(nigtype + nbgtype):
        key = "initial_gases" if i < nigtype else "injection_gases"

        # Record 4
        data = _nextsplitline(f, 1, skip_empty=True, comments="#")
        # igtype = int(data[0])

        tmp = []

        # Record 6
        line = _nextline(f, skip_empty=True, comments="#")
        while not line.replace("'", "").startswith("*"):
            data = _nextsplitline(line, 2)
            tmp2 = {"name": data[0]}
            if key == "initial_gases":
                tmp2["partial_pressure"] = to_float(data[1])

            else:
                tmp2["mole_fraction"] = to_float(data[1])

            tmp.append(tmp2)
            line = _nextline(f).strip()

        zones[key].append(tmp)

    return zones


def _read_zppr(f):
    """Read permeability-porosity law zones."""
    zones = {"permeability_porosity": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Record 3
    nppzone = int(line.strip())

    # Loop nppzone times
    for _ in range(nppzone):
        # Record 4
        data = _nextsplitline(f, 1, skip_empty=True, comments="#")
        # ippzone = int(data[0])

        # Record 6
        # Here, '*' does not mark the end of the list
        data = _nextsplitline(f, 3, skip_empty=True, comments="#")
        tmp = {
            "id": int(data[0]),
            "a": to_float(data[1]),
            "b": to_float(data[2]),
        }

        zones["permeability_porosity"].append(tmp)

    # '*' is not used here to mark the end of the list. Skip it.
    _ = f.next()

    return zones


def _read_zads(f):
    """Read surface adsorption zones."""
    zones = {"adsorption": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Record 3
    ndtype = int(line.strip())

    # Loop ndtype times
    for _ in range(ndtype):
        # Record 5
        data = _nextsplitline(f, 2, skip_empty=True, comments="#")
        # idzone = int(data[0])

        tmp = {
            "flag": int(data[1]),
            "species": [],
        }

        # Record 6
        # Not sure about the parsing, the manual states that '*' is only
        # written once in record 7 at the end of the zone list
        # If it's the case, it's unclear how the code knows how many
        # records to read per zone
        line = _nextline(f, skip_empty=True, comments="#")
        while not line.replace("'", "").startswith("*"):
            data = _nextsplitline(line, 3)
            tmp2 = {
                "name": data[0],
                "area_unit": int(data[1]),
                "area": to_float(data[2]),
            }

            tmp["species"].append(tmp2)
            line = _nextline(f).strip()

        zones["adsorption"].append(tmp)

    # '*' is not used here to mark the end of the list. Skip it.
    # _ = f.next()

    return zones


def _read_zlkd(f):
    """Read linear Kd zones."""
    zones = {"linear_kd": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Record 3
    kdtype = int(line.strip())

    # Loop kdtype times
    for _ in range(kdtype):
        # Record 4
        data = _nextsplitline(f, 1, skip_empty=True, comments="#")
        # idtype = int(data[0])

        tmp = []

        # Record 6
        # Not sure about the parsing, the manual states that '*' is only
        # written once in record 7 at the end of the zone list
        # If it's the case, it's unclear how the code knows how many
        # records to read per zone
        line = _nextline(f, skip_empty=True, comments="#")
        while not line.replace("'", "").startswith("*"):
            data = _nextsplitline(line, 3)
            tmp2 = {
                "name": data[0],
                "solid_density": to_float(data[1]),
                "value": to_float(data[2]),
            }

            tmp.append(tmp2)
            line = _nextline(f).strip()

        zones["linear_kd"].append(tmp)

    # '*' is not used here to mark the end of the list. Skip it.
    # _ = f.next()

    return zones


def _read_zexc(f, nxsites):
    """Read cation exchange zones."""
    zones = {"cation_exchange": []}

    # Find next non skip record
    line = _nextline(f, skip_empty=True, comments="#")

    if line.startswith("*"):
        return {}

    # Record 3
    nxtype = int(line.strip())

    # Loop nxtype times
    for _ in range(nxtype):
        # Record 5
        data = _nextsplitline(f, nxsites + 1, skip_empty=True, comments="#")
        # ixtype = int(data[0])

        tmp = [to_float(x) for x in data[1:]]
        zones["cation_exchange"].append(tmp)

    # '*' is not used here to mark the end of the list. Skip it.
    _ = f.next()

    return zones


def _nextline(f, remove_quotes=True, **kwargs):
    """
    Go to next line.

    Note
    ----
    Remove trailing Fortran comments.

    """
    line = f.next(**kwargs)
    line = line.replace('"', "'")

    # Remove quotes
    if remove_quotes:
        line = line.replace("'", "")

    # Remove Fortran comments
    if "!" in line:
        line = line[: line.index("!")]

    return line


def _nextsplitline(f_or_line, n=0, **kwargs):
    """Go to next line and split it."""
    if not isinstance(f_or_line, str):
        line = _nextline(f_or_line, **kwargs).strip()
    else:
        line = f_or_line

    data = line.split()
    if len(data) < n:
        raise ReadError(f"expected at least {n} entries, got {len(data)}.")

    return data



from ...._common import open_file
from ...output.tecplot._tecplot import read_buffer


def read(filename):
    """
    Read TECPLOT table file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Input file name or buffer.

    Returns
    -------
    dict
        Table data.

    """
    with open_file(filename, "r") as f:
        headers, zones = read_buffer(f)

    out = {headers[0]: zones[0]["data"][:, 0]}
    for zone in zones:
        out[zone["title"]] = zone["data"][:, 1]

    return out



sections = [
    "title",
    "options",
    "filenames",
    "weights_coefficients",
    "convergence",
    "output",
    "elements",
    "components",
    "minerals",
    "aqueous_species",
    "surface_complexes",
    "exchange_species",
    "default",
    "zones",
    "convergence_bounds",
    "end",
    "end_comments",
]



from ..._common import block_to_format, str2format
from ..._io.input.tough._helpers import write_record


def block(keyword):
    """Decorate block writing functions."""

    def decorator(func):
        from functools import wraps

        header = "----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8"

        @wraps(func)
        def wrapper(f, *args):
            f.write(f"{keyword}{header}\n")
            func(f, *args)
            f.write("\n")

        return wrapper

    return decorator


def _write_eleme(labels, materials, volumes, nodes, material_name=None):
    """Return a generator that iterates over the records of block ELEME."""
    label_length = len(labels[0])
    fmt = block_to_format["ELEME"][label_length]
    fmt = str2format(fmt)

    iterables = zip(labels, materials, volumes, nodes)
    for label, material, volume, node in iterables:
        mat = (
            material_name[material]
            if material_name and material in material_name
            else material
        )
        mat = mat if isinstance(mat, str) else f"{str(mat):>5}"
        record = write_record(
            [
                label,  # ID
                None,  # NSEQ
                None,  # NADD
                mat,  # MAT
                volume,  # VOLX
                None,  # AHTX
                None,  # PMX
                node[0],  # X
                node[1],  # Y
                node[2],  # Z
            ],
            fmt=fmt,
        )

        yield record[0]


def _write_coord(nodes):
    """Return a generator that iterates over the records of block COORD."""
    fmt = block_to_format["COORD"]
    fmt = str2format(fmt)

    for node in nodes:
        record = write_record(node, fmt)

        yield record[0]


def _write_conne(clabels, isot, d1, d2, areas, angles):
    """Return a generator that iterates over the records of block CONNE."""
    label_length = len(clabels[0][0])
    fmt = block_to_format["CONNE"][label_length]
    fmt = str2format(fmt)

    iterables = zip(clabels, isot, d1, d2, areas, angles)
    for label, isot, d1, d2, area, angle in iterables:
        record = write_record(
            [
                "".join(label),  # ID1-ID2
                None,  # NSEQ
                None,  # NAD1
                None,  # NAD2
                isot,  # ISOT
                d1,  # D1
                d2,  # D2
                area,  # AREAX
                angle,  # BETAX
                None,  # SIGX
            ],
            fmt=fmt,
        )

        yield record[0]


def _write_incon(
    labels, values, porosity=None, userx=None, phase_composition=None, eos=None
):
    """Return a generator that iterates over the records of block INCON."""
    porosity = porosity if porosity is not None else [None] * len(labels)
    userx = userx if userx is not None else [None] * len(labels)
    phase_composition = (
        phase_composition if phase_composition is not None else [None] * len(labels)
    )
    label_length = len(labels[0])
    fmt = block_to_format["INCON"]
    fmt1 = str2format(
        fmt[eos][label_length] if eos in fmt else fmt["default"][label_length]
    )
    fmt2 = str2format(fmt[0])

    iterables = zip(labels, values, porosity, userx, phase_composition)
    for label, value, phi, usrx, indicat0 in iterables:
        value = [v if v > -1.0e9 else None for v in value]
        cond1 = any(v is not None for v in value)
        cond2 = phi is not None
        cond3 = usrx is not None

        if cond1 or cond2 or cond3:
            # Record 1
            values = [
                label,
                None,
                None,
                phi,
            ]

            if eos == "tmvoc":
                values += [indicat0]

            else:
                values += list(usrx) if usrx is not None else []

            record = write_record(values, fmt1)[0]

            # Record 2
            record += write_record(value, fmt2, multi=True)[0]

            yield record

        else:
            continue



"""
Plot profiles
=============

The objective of this example is to import the output results and plot the profiles of temperature, pressure, liquid saturation and air mass fraction.

"""

########################################################################################
# Here, we assume that the simulation's output have been written in the file "OUTPUT". To import the results, we use the function :func:`toughio.read_output`. The variable `outputs` is a list with three :class:`toughio.Output` corresponding to the three time steps requested in the preprocessing example. In this example, we want to look at the last time step (index -1).

import numpy as np
import toughio

outputs = toughio.read_output("OUTPUT")
output = outputs[-1]

t = output.data["T"]
p = output.data["P"]
sl = output.data["SL"]
xm = output.data["XAIRG"]

########################################################################################

########################################################################################
# It is well known that for the stated conditions (1-D radial geometry, homogeneous medium, uniform initial conditions, and a constant-rate line source) the problem has a similarity solution: The partial differential equations for this complex two-phase flow problem can be rigorously transformed into a set of ordinary differential equations in the variable :math:`Z = R/\sqrt{t}`, which can be easily solved to any degree of accuracy desired by means of one-dimensional numerical integration (O'Sullivan, 1981). Comparison of TOUGH2 simulations with the semi-analytical similarity solution has shown excellent agreement (Doughty and Pruess, 1992). To define such variable, we first need to import the mesh 

mesh = toughio.read_mesh("mesh.pickle")
R = np.log(mesh.centers[:, 0] / (output.time) ** 0.5)

########################################################################################

########################################################################################
# Now that the required data have been imported, we can plot the results using :mod:`matplotlib`.

import matplotlib.pyplot as plt
plt.rc("font", size=12)

fig, ax1 = plt.subplots(figsize=(8, 5))
ax2 = ax1.twinx()

ax1.plot(R, t, color="black", linestyle="--", label="Temperature")
ax1.set_ylim(0.0, 260.0)
ax1.set_ylabel("Temperature ($\degree$C)")

ax2.plot(R, sl, label="Liquid saturation")
ax2.plot(R, xm, label="Air mass fraction")
ax2.plot(R, p * 1.0e-5, label="Pressure (bar)")
ax2.set_ylim(0.0, 1.3)
ax2.set_ylabel("Liquid saturation, air mass fraction, pressure")

ax1.set_xlim(R.min(), -4.0)
ax1.set_xlabel("$ln(R/\sqrt{t})$")

fig.legend(
    loc="lower left",
    bbox_to_anchor=(0.0, 0.0),
    bbox_transform=ax1.transAxes,
    frameon=False,
)

########################################################################################



"""
Generate mesh in Python with Gmsh
=================================

This examples describes how to generate a mesh using Gmsh in Python. It requires :mod:`pygmsh` v7 to be installed:

.. code-block:: bash

    pip install pygmsh --user

The mesh can also be generated from scratch directly in Gmsh either using its GUI and/or its internal scripting language.

Note that :mod:`toughio` is not required at this preliminary stage of the pre-processing. This example only intends to show how the mesh used in this sample problem has been generated. The user is free to use any meshing software as long as the mesh format is supported by :mod:`toughio` (through :mod:`meshio`).

"""

########################################################################################
# First, we import :mod:`numpy` and :mod:`pygmsh`.

import numpy as np
import pygmsh

########################################################################################

########################################################################################
# We can define a bunch of useful variables such as the characteristic length or some parameters to characterize the model.

lc = 100.0                       # Characteristic length of mesh
xmin, xmax = 0.0, 2000.0         # X axis boundaries
zmin, zmax = -500.0, -2500.0     # Z axis boundaries

inj_z = -1500.0                  # Depth of injection
flt_offset = 50.0                # Offset of fault
flt_thick = 10.0                 # Thickness of fault
tana = np.tan(np.deg2rad(80.0))  # Tangeant of dipping angle of fault
dist = 500.0 - 0.5 * flt_thick   # Distance from injection point (0.0, -1500.0) to left wall of fault

bnd_thick = 10.0                 # Thickness of boundary elements

########################################################################################

########################################################################################
# We start by defining the geometrical entity representing the fault.
# The fault is represented as a finite-thickness element that intersects all the layers of the model.
# To ensure conformity of the final mesh, each wall of the fault is represented by a segmented line where the positions of the nodes correspond to the intersections of the hanging (left) and foot (right) walls with the different layers.
# Note that the foot wall is inverted so that the fault entity forms a closed rectangular loop.

depths = [zmin, -1300.0, -1450.0, -1550.0, -1700.0, zmax]
fault_left = [[dist + (z - inj_z) / tana, z, 0.0] for z in depths]

depths = [zmin, -1300.0 + flt_offset, -1450.0 + flt_offset, -1550.0 + flt_offset, -1700.0 + flt_offset, zmax]
fault_right = [[dist + (z - inj_z) / tana + flt_thick, z, 0.0] for z in depths]

fault_pts = fault_left + fault_right[::-1]

########################################################################################

########################################################################################
# Now, we define the aquifer located at the left side of the fault.

cenaq_left_pts = [
    [xmin, -1450.0, 0.0],
    [dist + (-1450.0 - inj_z) / tana, -1450.0, 0.0],
    [dist + (-1550.0 - inj_z) / tana, -1550.0, 0.0],
    [xmin, -1550.0, 0.0],
]
capro_top_left_pts = [
    [xmin, -1300.0, 0.0],
    [dist + (-1300.0 - inj_z) / tana, -1300.0, 0.0],
    [dist + (-1450.0 - inj_z) / tana, -1450.0, 0.0],
    [xmin, -1450.0, 0.0],
]
capro_bot_left_pts = [
    [xmin, -1550.0, 0.0],
    [dist + (-1550.0 - inj_z) / tana, -1550.0, 0.0],
    [dist + (-1700.0 - inj_z) / tana, -1700.0, 0.0],
    [xmin, -1700.0, 0.0],
]
uppaq_left_pts = [
    [xmin, zmin, 0.0],
    [dist + (zmin - inj_z) / tana, zmin, 0.0],
    [dist + (-1300.0 - inj_z) / tana, -1300.0, 0.0],
    [xmin, -1300.0, 0.0],
]
basaq_left_pts = [
    [xmin, -1700.0, 0.0],
    [dist + (-1700.0 - inj_z) / tana, -1700.0, 0.0],
    [dist + (zmax - inj_z) / tana, zmax, 0.0],
    [xmin, zmax, 0.0],
]

########################################################################################

########################################################################################
# Likewise, we also define the aquifer located at the right side of the fault.

cenaq_right_pts = [
    [dist + (-1450.0 - inj_z + flt_offset) / tana + flt_thick, -1450.0 + flt_offset, 0.0],
    [xmax, -1450.0 + flt_offset, 0.0],
    [xmax, -1550.0 + flt_offset, 0.0],
    [dist + (-1550.0 - inj_z + flt_offset) / tana + flt_thick, -1550.0 + flt_offset, 0.0],
]
capro_top_right_pts = [
    [dist + (-1300.0 - inj_z + flt_offset) / tana + flt_thick, -1300.0 + flt_offset, 0.0],
    [xmax, -1300.0 + flt_offset, 0.0],
    [xmax, -1450.0 + flt_offset, 0.0],
    [dist + (-1450.0 - inj_z + flt_offset) / tana + flt_thick, -1450.0 + flt_offset, 0.0],
]
capro_bot_right_pts = [
    [dist + (-1550.0 - inj_z + flt_offset) / tana + flt_thick, -1550.0 + flt_offset, 0.0],
    [xmax, -1550.0 + flt_offset, 0.0],
    [xmax, -1700.0 + flt_offset, 0.0],
    [dist + (-1700.0 - inj_z + flt_offset) / tana + flt_thick, -1700.0 + flt_offset, 0.0],
]
uppaq_right_pts = [
    [dist + (zmin - inj_z) / tana + flt_thick, zmin, 0.0],
    [xmax, zmin, 0.0],
    [xmax, -1300.0 + flt_offset, 0.0],
    [dist + (-1300.0 - inj_z + flt_offset) / tana + flt_thick, -1300.0 + flt_offset, 0.0],
]
basaq_right_pts = [
    [dist + (-1700.0 - inj_z + flt_offset) / tana + flt_thick, -1700.0 + flt_offset, 0.0],
    [xmax, -1700.0 + flt_offset, 0.0],
    [xmax, zmax, 0.0],
    [dist + (zmax-inj_z) / tana + flt_thick, zmax, 0.0],
]

########################################################################################

########################################################################################
# Then, we define the boundary elements.
# In this sample problem, a no-flow boundary condition is imposed on the left side of the model (default in TOUGH), and Dirichlet boundary conditions are imposed elsewhere.
# Thus, physical boundary elements must be defined at the top, right and bottom sides of the model.
# Similarly to the fault entity, boundary entities are segmented to ensure conformity of the final mesh.

bound_right_pts = [
    [xmax, zmin, 0.0],
    [xmax + bnd_thick, zmin, 0.0],
    [xmax + bnd_thick, zmax, 0.0],
    [xmax, zmax, 0.0],
    [xmax, -1700.0 + flt_offset, 0.0],
    [xmax, -1550.0 + flt_offset, 0.0],
    [xmax, -1450.0 + flt_offset, 0.0],
    [xmax, -1300.0 + flt_offset, 0.0],
]
bound_top_pts = [
    [xmin, zmin, 0.0],
    [dist + (zmin - inj_z) / tana, zmin, 0.0],
    [dist + (zmin - inj_z) / tana + flt_thick, zmin, 0.0],
    [xmax, zmin, 0.0],
    [xmax + bnd_thick, zmin, 0.0],
    [xmax + bnd_thick, zmin + bnd_thick, 0.0],
    [dist + (zmin - inj_z + bnd_thick) / tana + flt_thick, zmin + bnd_thick, 0.0],
    [dist + (zmin - inj_z + bnd_thick) / tana, zmin + bnd_thick, 0.0],
    [xmin, zmin + bnd_thick, 0.0],
]
bound_bot_pts = [
    [xmin, zmax, 0.0],
    [dist + (zmax - inj_z) / tana, zmax, 0.0],
    [dist + (zmax - inj_z) / tana + flt_thick, zmax, 0.0],
    [xmax, zmax, 0.0],
    [xmax + bnd_thick, zmax, 0.0],
    [xmax + bnd_thick, zmax - bnd_thick, 0.0],
    [dist + (zmax - inj_z - bnd_thick) / tana + flt_thick, zmax - bnd_thick, 0.0],
    [dist + (zmax - inj_z - bnd_thick) / tana, zmax - bnd_thick, 0.0],
    [xmin, zmax - bnd_thick, 0.0],
]

########################################################################################

########################################################################################
# Once all the points have been created, we can now generate the geometry, assign rock types/materials as Gmsh physical properties, and generate the mesh.
# To refine the mesh in the injection zone, the characteristic length of each layer entity is increased the farther we get from the injection point.
# Note that layers are defined such that their characteristic lengths are increasing. This is because Gmsh keeps the first node defined in the geometry in case it detects duplicated nodes.

with pygmsh.geo.Geometry() as geo:
    # Define polygons
    fault = geo.add_polygon(fault_pts, mesh_size=0.1 * lc)
    cenaq_left = geo.add_polygon(cenaq_left_pts, mesh_size=0.1 * lc)
    capro_top_left = geo.add_polygon(capro_top_left_pts, mesh_size=0.2 * lc)
    capro_bot_left = geo.add_polygon(capro_bot_left_pts, mesh_size=0.2 * lc)
    uppaq_left = geo.add_polygon(uppaq_left_pts, mesh_size=2.0 * lc)
    basaq_left = geo.add_polygon(basaq_left_pts, mesh_size=2.0 * lc)
    cenaq_right = geo.add_polygon(cenaq_right_pts, mesh_size=0.75 * lc)
    capro_top_right = geo.add_polygon(capro_top_right_pts, mesh_size=0.75 * lc)
    capro_bot_right = geo.add_polygon(capro_bot_right_pts, mesh_size=0.75 * lc)
    uppaq_right = geo.add_polygon(uppaq_right_pts, mesh_size=2.0 * lc)
    basaq_right = geo.add_polygon(basaq_right_pts, mesh_size=2.0 * lc)
    bound_right = geo.add_polygon(bound_right_pts, mesh_size=lc)
    bound_top = geo.add_polygon(bound_top_pts, mesh_size=lc)
    bound_bot = geo.add_polygon(bound_bot_pts, mesh_size=lc)

    # Define materials
    geo.add_physical([uppaq_left, uppaq_right], "UPPAQ")
    geo.add_physical([capro_top_left, capro_bot_left, capro_top_right, capro_bot_right], "CAPRO")
    geo.add_physical([cenaq_left, cenaq_right], "CENAQ")
    geo.add_physical([basaq_left, basaq_right], "BASAQ")
    geo.add_physical(fault, "FAULT")
    geo.add_physical([bound_right, bound_top, bound_bot], "BOUND")

    # Remove duplicate entities
    geo.env.removeAllDuplicates()
    mesh = geo.generate_mesh(dim=2, algorithm=6)

    # Convert cell sets to material
    cell_data = [np.empty(len(c.data), dtype=int) for c in mesh.cells]
    field_data = {}
    for i, (k, v) in enumerate(mesh.cell_sets.items()):
        if k:
            field_data[k] = np.array([i + 1, 3])
            for ii, vv in enumerate(v):
                cell_data[ii][vv] = i + 1
    mesh.cell_data["material"] = cell_data
    mesh.field_data.update(field_data)
    mesh.cell_sets = {}

    # Remove lower dimension entities
    idx = [i for i, cell in enumerate(mesh.cells) if cell.type == "triangle"]
    mesh.cells = [mesh.cells[i] for i in idx]
    mesh.cell_data = {k: [v[i] for i in idx] for k, v in mesh.cell_data.items()}

    # Export the mesh for post-processing
    mesh.write("mesh.vtu")

########################################################################################

########################################################################################
# The generated mesh can be visualized in Python with :mod:`pyvista`.

import pyvista
pyvista.set_plot_theme("document")

p = pyvista.Plotter(window_size=(800, 800))
p.add_mesh(
    mesh=pyvista.from_meshio(mesh),
    scalar_bar_args={"title": "Materials"},
    show_scalar_bar=True,
    show_edges=True,
)
p.view_xy()
p.show()

########################################################################################



from ...._common import open_file
from .._common import getval, write_ffrecord
from ._helpers import section

__all__ = [
    "write",
]


def write(filename, parameters, verbose=True):
    """
    Write TOUGHREACT chemical input file.

    Parameters
    ----------
    filename : str, pathlike or buffer
        Output file name or buffer.
    parameters : dict
        Parameters to export.
    verbose : bool, optional, default True
        If `True`, add comments to describe content of file.

    """
    buffer = write_buffer(parameters, verbose)
    with open_file(filename, "w") as f:
        for record in buffer:
            f.write(record)


def write_buffer(parameters, verbose):
    """Write TOUGHREACT chemical input file."""
    # Define input file contents
    out = []
    out += _write_title(parameters)
    out += ["# Definition of the geochemical system"] if verbose else []
    out += _write_prim(parameters, verbose)
    out += _write_akin(parameters, verbose)
    out += _write_aque(parameters)
    out += _write_miner(parameters, verbose)
    out += _write_gas(parameters, verbose)
    out += _write_surx(parameters)
    out += _write_kdde(parameters, verbose)
    out += _write_exch(parameters, verbose)
    out += _write_water(parameters, verbose)
    out += _write_imin(parameters, verbose)
    out += _write_igas(parameters, verbose)
    out += _write_zppr(parameters, verbose)
    out += _write_zads(parameters, verbose)
    out += _write_zlkd(parameters, verbose)
    out += _write_zexc(parameters, verbose)
    out += _write_end_comments(parameters) if "end_comments" in parameters else []

    return "\n".join(out)


@section("# Title", f"#{'-' * 79}")
def _write_title(parameters):
    """Write title."""
    return [getval(parameters, "title", "")]


@section("# Primary aqueous species")
def _write_prim(parameters, verbose):
    """Write primary species."""
    out = []

    if not ("primary_species" in parameters and parameters["primary_species"]):
        return out

    if verbose:
        item = max(parameters["primary_species"], key=lambda x: len(x["name"]))
        fmt = f"{{:{min(len(item['name']), 20)}}}"

    for specie in parameters["primary_species"]:
        values = [
            getval(specie, "name", "''"),
            getval(specie, "transport", 0),
        ]

        if values[-1] == 2:
            values += [
                getval(specie, "mineral_name", "''"),
                getval(specie, "sorption_density", 0.0),
                getval(specie, "adsorption_id", 0),
            ]

            if values[-1] == 1:
                values += [getval(specie, "capacitance", 0.0)]

        out += write_ffrecord(values, verbose, str_fmt=fmt if verbose else None)

    return out


@section("# Aqueous kinetics")
def _write_akin(parameters, verbose):
    """Write aqueous kinetics."""
    out = []

    if not ("aqueous_kinetics" in parameters and parameters["aqueous_kinetics"]):
        return out

    # Record 2
    out += [f"{len(parameters['aqueous_kinetics'])}"]

    for i, kinetic in enumerate(parameters["aqueous_kinetics"]):
        # Record 3
        out += [f"{i + 1}"]

        # Record 4
        n = len(kinetic["species"])
        values = [n]
        if n:
            for specie in kinetic["species"]:
                values += [
                    getval(specie, "stoichiometric_coeff", 1.0),
                    getval(specie, "name", "''"),
                ]

        if "reaction_affinity" in kinetic:
            values += [
                getval(kinetic, ("reaction_affinity", "id"), 0),
                getval(kinetic, ("reaction_affinity", "cf"), 0.0),
                getval(kinetic, ("reaction_affinity", "logK"), 0.0),
            ]

        out += write_ffrecord(values)

        # Record 5
        values = [
            getval(kinetic, "id", 0),
            getval(kinetic, "n_mechanism", 0),
        ]
        tmp = write_ffrecord(values)

        # Record 6
        rate = getval(kinetic, "rate", 0.0)
        rate = rate if not isinstance(rate, dict) else -1.0
        tmp += [f"{rate}"]

        # Record 6.1
        if rate == -1.0:
            values = [
                getval(kinetic, ("rate", "k25"), 0.0),
                getval(kinetic, ("rate", "Ea"), 0.0),
            ]
            tmp += write_ffrecord(values)

        # Record 7
        ncp_rx1 = len(kinetic["product"])
        values = [ncp_rx1]
        if ncp_rx1:
            for specie in kinetic["product"]:
                values += [
                    getval(specie, "specie", "''"),
                    getval(specie, "flag", 0),
                    getval(specie, "power", 0.0),
                ]
        tmp += write_ffrecord(values)

        # Record 8
        ncp_rx2 = len(kinetic["monod"])
        values = [ncp_rx2]
        if ncp_rx2:
            for specie in kinetic["monod"]:
                values += [
                    getval(specie, "specie", "''"),
                    getval(specie, "flag", 0),
                    getval(specie, "half_saturation", 0.0),
                ]
        tmp += write_ffrecord(values)

        # Record 9
        ncp_rx3 = len(kinetic["inhibition"])
        values = [ncp_rx3]
        if ncp_rx3:
            for specie in kinetic["inhibition"]:
                values += [
                    getval(specie, "specie", "''"),
                    getval(specie, "flag", 0),
                    getval(specie, "constant", 0.0),
                ]
        tmp += write_ffrecord(values)

        # Write records 5 to 9
        if verbose:
            n = max(len(max(tmp, key=len)), 30)
            tmp[0] = f"{tmp[0]:<{n}}  ! Kinetic model index, number of mechanisms"
            tmp[1] = f"{tmp[1]:<{n}}  ! Forward rate constant"
            tmp[-3] = f"{tmp[-3]:<{n}}  ! Species in product term"
            tmp[-2] = f"{tmp[-2]:<{n}}  ! Species in monod terms"
            tmp[-1] = f"{tmp[-1]:<{n}}  ! Species in inhibition terms"

        out += tmp

    return out


@section("# Aqueous complexes")
def _write_aque(parameters):
    """Write aqueous species."""
    out = []

    if not ("aqueous_species" in parameters and parameters["aqueous_species"]):
        return out

    for specie in parameters["aqueous_species"]:
        out += [f"{specie:<20}"]

    return out


@section("# Minerals")
def _write_miner(parameters, verbose):
    """Write minerals."""
    out = []

    if not ("minerals" in parameters and parameters["minerals"]):
        return out

    if verbose:
        item = max(parameters["minerals"], key=lambda x: len(x["name"]))
        fmt = f"{{:{min(len(item['name']), 20)}}}"

    for mineral in parameters["minerals"]:
        # Record 2
        ikin = getval(mineral, "type", 0)
        idispre = getval(mineral, "kinetic_constraint", 0)
        values = [
            getval(mineral, "name", "''"),
            ikin,
            idispre,
            getval(mineral, "solid_solution", 0),
            getval(mineral, "precipitation_dry", 0),
        ]
        out += write_ffrecord(values, verbose, str_fmt=fmt if verbose else None)

        # Records 2.1 and 2.2
        if ikin == 1:
            n = 2 if idispre == 3 else 1
            for ir in range(n):
                if idispre == 1:
                    key = "dissolution"

                elif idispre == 2:
                    key = "precipitation"

                elif idispre == 3:
                    key = "dissolution" if ir == 0 else "precipitation"

                idep = getval(mineral, (key, "rate_ph_dependence"), 0)
                values = [
                    getval(mineral, (key, "k25"), 0.0),
                    idep,
                    getval(mineral, (key, "eta"), 0.0),
                    getval(mineral, (key, "theta"), 0.0),
                    getval(mineral, (key, "activation_energy"), 0.0),
                    getval(mineral, (key, "a"), 0.0),
                    getval(mineral, (key, "b"), 0.0),
                    getval(mineral, (key, "c"), 0.0),
                ]

                # Record 2.2
                if key == "precipitation":
                    values += [
                        getval(mineral, (key, "volume_fraction_ini"), 0.0),
                        getval(mineral, (key, "id"), 0),
                    ]

                out += write_ffrecord(values)

                # Record 2.1.1
                if idep == 1:
                    values = [
                        getval(mineral, (key, "ph1"), 0.0),
                        getval(mineral, (key, "slope1"), 0.0),
                        getval(mineral, (key, "ph2"), 0.0),
                        getval(mineral, (key, "slope2"), 0.0),
                    ]
                    out += write_ffrecord(values)

                # Record 2.1.2
                elif idep == 2:
                    ndis = (
                        len(mineral[key]["extra_mechanisms"])
                        if "extra_mechanisms" in mineral[key]
                        else 0
                    )
                    values = [ndis]
                    out += write_ffrecord(values)

                    # Record 2.1.2.1
                    if ndis:
                        for mechanism in mineral[key]["extra_mechanisms"]:
                            values = [
                                getval(mechanism, "ki", 0.0),
                                getval(mechanism, "activation_energy", 0.0),
                            ]

                            nspds = len(getval(mechanism, "species", []))
                            values += [nspds]
                            if nspds:
                                for specie in mechanism["species"]:
                                    values += [
                                        getval(specie, "name", "''"),
                                        getval(specie, "power", 0.0),
                                    ]
                            out += write_ffrecord(values)

        # Record 2.3
        if idispre != 1:
            values = [
                getval(mineral, "gap", 0.0),
                getval(mineral, "temp1", 0.0),
                getval(mineral, "temp2", 0.0),
            ]
            out += write_ffrecord(values)

    return out


@section("# Gases")
def _write_gas(parameters, verbose):
    """Write gaseous species."""
    out = []

    if not ("gaseous_species" in parameters and parameters["gaseous_species"]):
        return out

    if verbose:
        item = max(parameters["gaseous_species"], key=lambda x: len(x["name"]))
        fmt = f"{{:{min(len(item['name']), 20)}}}"

    for specie in parameters["gaseous_species"]:
        values = [
            getval(specie, "name", "''"),
            getval(specie, "fugacity", 0),
        ]
        out += write_ffrecord(values, verbose, str_fmt=fmt if verbose else None)

    return out


@section("# Surface complexes")
def _write_surx(parameters):
    """Write surface complexes."""
    out = []

    if not ("surface_complexes" in parameters and parameters["surface_complexes"]):
        return out

    for complex_ in parameters["surface_complexes"]:
        out += [f"{complex_:<20}"]

    return out


@section("# Species with Kd and decay")
def _write_kdde(parameters, verbose):
    """Write Kd and decay."""
    out = []

    if not ("kd_decay" in parameters and parameters["kd_decay"]):
        return out

    if verbose:
        item = max(parameters["kd_decay"], key=lambda x: len(x["name"]))
        fmt = f"{{:{min(len(item['name']), 20)}}}"

    for specie in parameters["kd_decay"]:
        values = [
            getval(specie, "name", "''"),
            getval(specie, "decay_constant", 0.0),
            getval(specie, "a", 0.0),
            getval(specie, "b", 0.0),
        ]
        out += write_ffrecord(values, verbose, str_fmt=fmt if verbose else None)

    return out


@section("# Exchangeable cations")
def _write_exch(parameters, verbose):
    """Write exchanges species."""
    out = []

    if not ("exchanged_species" in parameters and parameters["exchanged_species"]):
        return out

    # Record 2
    nxsites = len(getval(parameters["exchanged_species"][0], "site_coeffs", []))
    values = [
        nxsites,
        getval(parameters, "exchange_sites_id", 0),
    ]
    out += write_ffrecord(values, verbose, int_fmt="{:<4d}")

    # Record 3
    if verbose:
        item = max(parameters["exchanged_species"], key=lambda x: len(x["name"]))
        n = min(len(item["name"]), 20)
        fmt = f"{{:{n}}}"

        out += [f"# {' ' * (n - 2)} ref. type    coeff."]

    for specie in parameters["exchanged_species"]:
        values = [
            getval(specie, "name", "''"),
            int(getval(specie, "reference", False)),
            getval(specie, "type", 1),
        ]
        values += [*getval(specie, "site_coeffs", [])]
        out += write_ffrecord(values, verbose, str_fmt=fmt if verbose else None)

    return out


@section("# Initial and injection water types")
def _write_water(parameters, verbose):
    """Write water zones."""
    out = []

    if "zones" not in parameters:
        return out

    nwtypes = {
        "initial_waters": 0,
        "injection_waters": 0,
    }
    for key in nwtypes:
        if key in parameters["zones"]:
            nwtypes[key] = len(parameters["zones"][key])

    if not sum(nwtypes.values()):
        return out

    values = list(nwtypes.values())
    out += write_ffrecord(values, verbose, int_fmt="{:<4d}")

    for k, v in nwtypes.items():
        if not v:
            continue

        for i, zone in enumerate(parameters["zones"][k]):
            # Record 4
            if verbose:
                out += ["#         T(C)    P(bar)"]

                if "rock" in zone:
                    out[-1] = f"{out[-1]}      Rock"

            values = [
                i + 1 if "rock" not in zone else -1,
                getval(zone, "temperature", 0.0),
                getval(zone, "pressure", 0.0),
            ]
            values += [getval(zone, "rock", "''")[:5]] if "rock" in zone else []
            out += write_ffrecord(values, verbose, int_fmt="{:<4d}", str_fmt="{:>9}")

            # Record 6
            if "species" in zone and zone["species"]:
                if verbose:
                    item = max(zone["species"], key=lambda x: len(x["name"]))
                    n = min(len(item["name"]), 20)
                    fmt1 = f"{{:{n}}}"

                    try:
                        item = max(zone["species"], key=lambda x: len(x["nameq"]))
                        fmt2 = f"{{:>{max(len(item['nameq']), 10)}}}"

                    except KeyError:
                        fmt2 = "{:>10}"

                    out += [
                        f"# {' ' * (n - 2)} icon     guess      ctot {fmt2.format('constraint')}  log(Q/K)"
                    ]

                for specie in zone["species"]:
                    name = getval(specie, "name", "''")
                    name = fmt1.format(name) if verbose else name
                    nameq = "*" if "nameq" not in specie else specie["nameq"]
                    values = [
                        getval(specie, "flag", 0),
                        getval(specie, "guess", 0.0),
                        getval(specie, "ctot", 0.0),
                        nameq,  # nameq is optional
                        getval(specie, "log_fugacity", 0.0),
                    ]
                    tmp = write_ffrecord(
                        values, verbose, str_fmt=fmt2 if verbose else None
                    )
                    out += [f"{name} {tmp[0]}"]

            # Record 7
            out += ["*"]

    return out[:-1]


@section("# Initial mineral zones")
def _write_imin(parameters, verbose):
    """Write mineral zones."""
    out = []

    if "zones" not in parameters:
        return out

    nmtype = 0
    if "minerals" in parameters["zones"]:
        nmtype = len(parameters["zones"]["minerals"])

    if not nmtype:
        return out

    values = [nmtype]
    out += write_ffrecord(values)

    for i, zone in enumerate(parameters["zones"]["minerals"]):
        # Record 4
        values = [i + 1] if "rock" not in zone else [-(i + 1), f"'{zone['rock'][:5]}'"]
        out += write_ffrecord(values)

        # Record 6
        if "species" in zone and zone["species"]:
            if verbose:
                item = max(zone["species"], key=lambda x: len(x["name"]))
                fmt = f"{{:{max(len(item['name']), 10)}}}"

            for mineral in zone["species"]:
                name = getval(mineral, "name", "''")
                name = fmt.format(name) if verbose else name
                flag = getval(mineral, "flag", 0)
                values = [
                    getval(mineral, "volume_fraction_ini", 0.0),
                    flag,
                ]
                tmp = write_ffrecord(values, verbose)
                out += [f"{name} {tmp[0]}"]

                # Record 6.1
                if flag == 1:
                    area_unit = getval(mineral, "area_unit", 0)
                    values = [
                        getval(mineral, "radius", 0.0),
                        getval(mineral, "area_ini", 0.0),
                        area_unit,
                    ]
                    out += write_ffrecord(values, verbose)
                    out[-1] = out[-1].lstrip()

                    # TODO: investigate what is the next record when IMFLG2 < 0
                    # This record is not documented in the user guide (p. 65)
                    if area_unit < 0:
                        values = getval(mineral, "unknown", [])
                        out += write_ffrecord(values, verbose)

        # Record 7
        out += ["*"]

    return out[:-1]


@section("# Initial and injection gas zones")
def _write_igas(parameters, verbose):
    """Write gas zones."""
    out = []

    if "zones" not in parameters:
        return out

    ngtypes = {
        "initial_gases": 0,
        "injection_gases": 0,
    }
    for key in ngtypes:
        if key in parameters["zones"]:
            ngtypes[key] = len(parameters["zones"][key])

    if not sum(ngtypes.values()):
        return out

    values = list(ngtypes.values())
    out += write_ffrecord(values, verbose, int_fmt="{:<4d}")

    for k, v in ngtypes.items():
        if not v:
            continue

        for i, zone in enumerate(parameters["zones"][k]):
            # Record 4
            values = [i + 1]
            out += write_ffrecord(values, verbose, int_fmt="{:<4d}")

            # Record 6
            if verbose:
                item = max(zone, key=lambda x: len(x["name"]))
                n = min(len(item["name"]), 20)
                fmt = f"{{:{n}}}"

                out += [
                    (
                        f"# {' ' * n} PP(bar)"
                        if k == "initial_gases"
                        else f"# {' ' * (n - 1)} PP(%mol)"
                    )
                ]

            for specie in zone:
                name = getval(specie, "name", "''")
                name = fmt.format(name) if verbose else name
                values = [
                    (
                        getval(specie, "partial_pressure", 0.0)
                        if k == "initial_gases"
                        else getval(specie, "mole_fraction", 0.0)
                    )
                ]
                tmp = write_ffrecord(values, verbose)
                out += [f"{name} {tmp[0]}"]

            # Record 7
            out += ["*"]

    return out[:-1]


@section("# Permeability-porosity zones")
def _write_zppr(parameters, verbose):
    """Write permeability-porosity law zones."""
    out = []

    if "zones" not in parameters:
        return out

    nppzone = 0
    if "permeability_porosity" in parameters["zones"]:
        nppzone = len(parameters["zones"]["permeability_porosity"])

    if not nppzone:
        return out

    values = [nppzone]
    out += write_ffrecord(values)

    for i, zone in enumerate(parameters["zones"]["permeability_porosity"]):
        # Record 4
        values = [i + 1]
        out += write_ffrecord(values)

        # Record 6
        if verbose:
            out += ["#        a-par     b-par"]

        values = [
            getval(zone, "id", 0),
            getval(zone, "a", 0.0),
            getval(zone, "b", 0.0),
        ]
        out += write_ffrecord(values, verbose, int_fmt="{:<4d}")

    return out


@section("# Initial surface adsorption zones")
def _write_zads(parameters, verbose):
    """Write surface adsorption zones."""
    out = []

    if "zones" not in parameters:
        return out

    ndtype = 0
    if "adsorption" in parameters["zones"]:
        ndtype = len(parameters["zones"]["adsorption"])

    if not ndtype:
        return out

    values = [ndtype]
    out += write_ffrecord(values)

    for i, zone in enumerate(parameters["zones"]["adsorption"]):
        # Record 5
        values = [
            i + 1,
            getval(zone, "flag", 0),
        ]
        out += write_ffrecord(values)

        # Record 6
        if "species" in zone and zone["species"]:
            if verbose:
                item = max(zone["species"], key=lambda x: len(x["name"]))
                n = max(len(item["name"]), 10)
                fmt = f"{{:{n}}}"

                out += [f"# {' ' * (n - 2)} unit      area"]

            for specie in zone["species"]:
                name = getval(specie, "name", "''")
                name = fmt.format(name) if verbose else name
                values = [
                    getval(specie, "area_unit", 0),
                    getval(specie, "area", 0.0),
                ]
                tmp = write_ffrecord(values, verbose)
                out += [f"{name} {tmp[0]}"]

        # Record 7
        out += ["*"]

    return out[:-1]


@section("# Initial linear equilibrium Kd zones")
def _write_zlkd(parameters, verbose):
    """Write linear Kd zones."""
    out = []

    if "zones" not in parameters:
        return out

    kdtype = 0
    if "linear_kd" in parameters["zones"]:
        kdtype = len(parameters["zones"]["linear_kd"])

    if not kdtype:
        return out

    values = [kdtype]
    out += write_ffrecord(values)

    for i, zone in enumerate(parameters["zones"]["linear_kd"]):
        # Record 4
        values = [i + 1]
        out += write_ffrecord(values)

        # Record 6
        if verbose:
            item = max(zone, key=lambda x: len(x["name"]))
            n = min(len(item["name"]), 10)
            fmt = f"{{:{n}}}"

            out += [f"# {' ' * n} density        Kd"]

        for specie in zone:
            name = getval(specie, "name", "''")
            name = fmt.format(name) if verbose else name
            values = [
                getval(specie, "solid_density", 0.0),
                getval(specie, "value", 0.0),
            ]
            tmp = write_ffrecord(values, verbose)
            out += [f"{name} {tmp[0]}"]

        # Record 7
        out += ["*"]

    return out[:-1]


@section("# Initial zones of cation exchange")
def _write_zexc(parameters, verbose):
    """Write cation exchange zones."""
    out = []

    if "zones" not in parameters:
        return out

    nxtype = 0
    if "cation_exchange" in parameters["zones"]:
        nxtype = len(parameters["zones"]["cation_exchange"])

    if not nxtype:
        return out

    values = [nxtype]
    out += write_ffrecord(values)

    if verbose:
        out += ["#     capacity"]

    for i, zone in enumerate(parameters["zones"]["cation_exchange"]):
        values = [i + 1, *zone]
        out += write_ffrecord(values, verbose, int_fmt="{:<4d}")

    return out


def _write_end_comments(parameters):
    """Write end comments."""
    end_comments = getval(parameters, "end_comments", "")

    return [end_comments] if isinstance(end_comments, str) else end_comments
